# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
import typing

from gi.repository import Adw, Gio, GObject, Gtk

import gsecrets.config_manager as config
from gsecrets.entry_row import EntryRow
from gsecrets.group_row import GroupRow
from gsecrets.safe_element import SafeElement, SafeGroup
from gsecrets.single_selection import SingleSelection
from gsecrets.sorting import SortingHat

if typing.TYPE_CHECKING:
    from gsecrets.safe_element import SafeEntry


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/browsing_panel.ui")
class BrowsingPanel(Adw.Bin):
    # pylint: disable=too-many-instance-attributes

    __gtype_name__ = "BrowsingPanel"

    _empty_group_box = Gtk.Template.Child()
    _empty_search_page = Gtk.Template.Child()
    _list_view = Gtk.Template.Child()
    _scrolled_window = Gtk.Template.Child()
    _stack = Gtk.Template.Child()

    _query = ""

    def __init__(self, unlocked_database):
        super().__init__()

        self._signals = GObject.SignalGroup.new(SafeGroup)
        self._signals.connect_closure(
            "children-changed",
            self._on_children_changed,
            False,
        )

        sorting = config.get_sort_order()
        self.entry_sorter = SortingHat.get_sorter(sorting)
        self.group_sorter = SortingHat.get_sorter(sorting)

        dbm = unlocked_database.database_manager

        self.current_group = dbm.root
        self._signals.props.target = self.current_group

        self.entries = Gtk.SortListModel.new(dbm.entries, self.entry_sorter)
        self.groups = Gtk.SortListModel.new(dbm.groups, self.group_sorter)

        self.unlocked_database = unlocked_database

        flatten = Gio.ListStore.new(Gtk.SortListModel)
        flatten.splice(0, 0, [self.groups, self.entries])

        flatten_model = Gtk.FlattenListModel.new(flatten)

        expr = Gtk.PropertyExpression.new(
            SafeElement.__gtype__, None, "parent_uuid_str"
        )
        self._filter = Gtk.StringFilter.new(expr)
        root_group = SafeGroup.get_root(dbm)
        self._filter.props.search = str(root_group.uuid)

        self._search_filter = Gtk.CustomFilter.new(self._search_filter_fn)

        self._list_model = Gtk.FilterListModel.new(flatten_model, self._filter)

        self.selection_model = SingleSelection(self._list_model)

        settings = unlocked_database.window.application.settings
        settings.connect("changed::sort-order", self._on_sort_order_changed)

        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_setup)
        factory.connect("bind", self._on_bind)

        self._list_view.set_model(self.selection_model)
        self._list_view.set_factory(factory)
        self._list_model.connect("notify::n-items", self._on_model_n_items_changed)
        self._set_stack_page()

        unlocked_database.database_manager.connect(
            "sorting_changed",
            self._on_sorting_changed,
        )

        self._list_view.connect("activate", self._on_listview_activate)
        self.selection_model.connect("notify::selected-item", self._on_selected_item)

    def _on_selected_item(self, model, _pspec):
        if element := model.props.selected_item:
            self.unlocked_database.show_edit_page(element)
        else:
            current = self.unlocked_database.current_element
            self.unlocked_database.show_browser_page(current.parentgroup)

    def visit_group(self, group):
        self.current_group = group
        self._signals.props.target = group
        self.unlocked_database.props.search_active = False
        self._filter.props.search = str(group.uuid)

    def set_search(self, query: str | None) -> None:
        if self._query == query:
            return

        if query is None:
            self._list_model.props.filter = self._filter
            self._set_stack_page()
            self._query = ""
            return

        self._list_model.props.filter = self._search_filter

        if query in self._query:
            change = Gtk.FilterChange.LESS_STRICT
        elif self._query in query:
            change = Gtk.FilterChange.MORE_STRICT
        else:
            change = Gtk.FilterChange.DIFFERENT

        self._query = query.lower()
        self._search_filter.changed(change)

    def hide_selection(self, hide):
        self.selection_model.props.hide_selected = hide

    def _on_listview_activate(self, _list_view, pos):
        element = self._list_model.get_item(pos)

        if element.is_group:
            self.unlocked_database.show_browser_page(element)
        elif self.unlocked_database.props.selection_mode:
            element.props.selected = not element.props.selected
        else:
            self.selection_model.props.selected_item = element
            self.unlocked_database.show_edit_page(element)

    def _on_setup(self, _list_view, item):
        entry_row = EntryRow(self.unlocked_database)

        item.props.child = entry_row

    def _on_bind(self, _list_view, item):
        element = item.props.item
        row = item.props.child
        if element.is_group:
            if row.__gtype_name__ != "GroupRow":
                row = GroupRow(self.unlocked_database)
                item.props.child = row

            row.props.safe_group = element
        else:
            if row.__gtype_name__ != "EntryRow":
                row = EntryRow(self.unlocked_database)
                item.props.child = row

            row.props.safe_entry = element

    def _on_sorting_changed(self, _db_manager, is_entry):
        if is_entry:
            self.entry_sorter.changed(Gtk.SorterChange.DIFFERENT)
        else:
            self.group_sorter.changed(Gtk.SorterChange.DIFFERENT)

    def _on_sort_order_changed(self, settings, _key):
        """Callback to be executed when the sorting has been changed."""
        sorting = settings.get_enum("sort-order")
        logging.debug("Sort order changed to %s", sorting)

        sorting = config.get_sort_order()
        self.entry_sorter = SortingHat.get_sorter(sorting)
        self.group_sorter = SortingHat.get_sorter(sorting)

        self.entries.set_sorter(self.entry_sorter)
        self.groups.set_sorter(self.group_sorter)

    def _set_stack_page(self):
        if not self._list_model.get_n_items():
            if self.unlocked_database.props.search_active:
                self._stack.props.visible_child = self._empty_search_page
            else:
                self._stack.props.visible_child = self._empty_group_box
        else:
            self._stack.props.visible_child = self._scrolled_window

    def _on_model_n_items_changed(self, _model, _pspec):
        self._set_stack_page()

    def _on_children_changed(self, _group, removed, added):
        if added > 0 and removed == 0:
            change = Gtk.FilterChange.LESS_STRICT
        elif removed > 0 and added == 0:
            change = Gtk.FilterChange.MORE_STRICT
        else:
            change = Gtk.FilterChange.DIFFERENT

        self._filter.changed(change)

    def _search_filter_fn(self, item: SafeEntry | SafeGroup) -> bool:
        if item.is_root_group:
            return False

        if not item.is_group and item.tags:
            for tag in item.tags:
                if self._query in tag.lower():
                    return True

        if isinstance(item, SafeGroup):
            fields = [item.name, item.notes]
        else:
            fields = [item.name, item.notes, item.url, item.username]

        for field in fields:
            if not field:  # type: ignore
                continue

            if self._query in field.lower():  # type: ignore
                return True

        return False

    def unselect(self, item):
        if item == self.selection_model.props.selected_item:
            self.selection_model.props.selected_item = None

    def select_next(self):
        pos = self.selection_model.selected_item_pos()
        n_items = self.selection_model.get_n_items()
        scroll_flags = Gtk.ListScrollFlags.NONE

        if pos == Gtk.INVALID_LIST_POSITION and n_items > 0:
            self.selection_model.props.selected_item = self.selection_model.get_item(0)
            self._list_view.scroll_to(0, scroll_flags, None)
            return

        if pos + 1 < n_items:
            self.selection_model.props.selected_item = self.selection_model.get_item(
                pos + 1,
            )
            self._list_view.scroll_to(pos + 1, scroll_flags, None)

    def select_previous(self):
        pos = self.selection_model.selected_item_pos()
        n_items = self.selection_model.get_n_items()
        scroll_flags = Gtk.ListScrollFlags.NONE

        if pos == Gtk.INVALID_LIST_POSITION and n_items > 0:
            self.selection_model.props.selected_item = self.selection_model.get_item(
                n_items - 1,
            )
            self._list_view.scroll_to(n_items - 1, scroll_flags, None)
            return

        if pos > 0:
            self.selection_model.props.selected_item = self.selection_model.get_item(
                pos - 1,
            )
            self._list_view.scroll_to(pos - 1, scroll_flags, None)



# SPDX-License-Identifier: GPL-3.0-only
# pylint: disable=too-many-lines
from __future__ import annotations

import binascii
import logging
import typing
from datetime import UTC, datetime
from enum import Enum
from gettext import gettext as _
from typing import NamedTuple

from gi.repository import Gio, GLib, GObject, Gtk
from pyotp import TOTP, parse_uri

from gsecrets.attributes_model import AttributesModel

if typing.TYPE_CHECKING:
    from uuid import UUID

    from pykeepass.attachment import Attachment
    from pykeepass.entry import Entry
    from pykeepass.group import Group

    from gsecrets.database_manager import (
        DatabaseManager,  # pylint: disable=ungrouped-imports
    )


class EntryColor(Enum):
    NONE = "NoneColorButton"
    BLUE = "BlueColorButton"
    GREEN = "GreenColorButton"
    YELLOW = "YellowColorButton"
    ORANGE = "OrangeColorButton"
    RED = "RedColorButton"
    PURPLE = "PurpleColorButton"
    BROWN = "BrownColorButton"

    def to_translatable(self):  # pylint: disable=too-many-return-statements
        match self:
            case EntryColor.NONE:
                return _("White")
            case EntryColor.BLUE:
                return _("Blue")
            case EntryColor.GREEN:
                return _("Green")
            case EntryColor.YELLOW:
                return _("Yellow")
            case EntryColor.ORANGE:
                return _("Orange")
            case EntryColor.RED:
                return _("Red")
            case EntryColor.PURPLE:
                return _("Purple")
            case EntryColor.BROWN:
                return _("Brown")

    def css_class(self):
        match self:
            case EntryColor.NONE:
                return "white"
            case EntryColor.BLUE:
                return "blue"
            case EntryColor.GREEN:
                return "green"
            case EntryColor.YELLOW:
                return "yellow"
            case EntryColor.ORANGE:
                return "orange"
            case EntryColor.RED:
                return "red"
            case EntryColor.PURPLE:
                return "purple"
            case EntryColor.BROWN:
                return "brown"


class SafeElement(GObject.Object):
    selected = GObject.Property(type=bool, default=False)
    sensitive = GObject.Property(type=bool, default=True)
    sorted_handler_id: int | None = None

    _notes = None
    _name = None

    def __init__(self, db_manager: DatabaseManager, element: Entry | Group):
        """GObject to handle a safe element.

        The underlying pykeepass element can be obtainied via the `element`
        property, when it is certain the element is a Group or Entry, the
        properties `entry` and `group` should be used instead.
        """
        super().__init__()

        self._element = element
        self._db_manager = db_manager

        self.is_group = isinstance(self, SafeGroup)
        self.is_entry = isinstance(self, SafeEntry)

    def __eq__(self, other):  # pylint: disable=arguments-differ
        if isinstance(other, SafeElement):
            return self.uuid == other.uuid

        return False

    def __hash__(self):  # pylint: disable=arguments-differ
        return hash(self.uuid)

    @GObject.Signal(flags=GObject.SignalFlags.ACTION)
    def updated(self):
        """Update the entry.

        Used to tell whenever there have been any changed that should be
        reflected on the main list box or edit page.
        """
        self._db_manager.is_dirty = True
        self.touch(modify=True)
        self.emit(self.updated)
        logging.debug("Safe element updated")

    def touch(self, modify: bool = False) -> None:
        """Update the last accessed time.

        If modify is true it also updates the last modified time.
        """
        self._element.touch(modify)

    def delete(self) -> None:
        """Delete an Element from the database."""
        element = self._element
        parentgroup = self.parentgroup

        self.delete_inner()

        if self.is_entry:
            self._db_manager.db.delete_entry(element)
        else:
            self._db_manager.db.delete_group(element)

        parentgroup.updated()

    def delete_inner(self) -> None:
        """Recursively removes items from our wrapper of the database."""
        if self.is_entry:
            found, pos = self._db_manager.entries.find(self)
            if found:
                self._db_manager.entries.remove(pos)
        else:
            while entry := self.entries.get_item(0):  # pylint: disable=no-member
                entry.delete_inner()

            while group := self.subgroups.get_item(0):  # pylint: disable=no-member
                group.delete_inner()

            found, pos = self._db_manager.groups.find(self)
            if found:
                self._db_manager.groups.remove(pos)

    def trash(self) -> bool:
        """Thrash an Element from the database.

        Return if the element was deleted instead of being sent to the trash
        bin.
        """
        element = self._element
        parentgroup = self.parentgroup

        if (trash_bin := self._db_manager.trash_bin) and (
            self._db_manager.parent_checker(self, trash_bin) and not self.is_trash_bin
        ):
            self.delete()
            return True

        if self.is_trash_bin:
            self.delete()
            self._db_manager.trash_bin = None
            return True

        trash_bin_missing = self._db_manager.trash_bin is None

        if self.is_entry:
            self._db_manager.db.trash_entry(element)
            parentgroup.filter_changed(True)
        else:
            self._db_manager.db.trash_group(element)
            parentgroup.filter_changed(False)

        parentgroup.emit_children_changed(1, 0)

        # We add the trash bin if it was not present already
        if trash_bin_missing and (
            trash_bin_inner := self._db_manager.db.recyclebin_group
        ):
            trash_bin = SafeGroup(self._db_manager, trash_bin_inner)
            self._db_manager.trash_bin = trash_bin
            self._db_manager.groups.append(trash_bin)
            trash_bin.parentgroup.updated()
            trash_bin.parentgroup.emit_children_changed(0, 1)

        if trash_bin := self._db_manager.trash_bin:
            trash_bin.filter_changed(self.is_entry)
            trash_bin.emit_children_changed(0, 1)

        parentgroup.updated()

        return False

    def move_to(self, dest: SafeGroup) -> None:
        old_location = self.parentgroup

        if old_location == dest:
            return

        # We signal items changed on the main list store so that the changes are
        # propagated to the filter models used by each group.
        if self.is_entry:
            self._db_manager.db.move_entry(self._element, dest.group)
            dest.filter_changed(True)
            old_location.filter_changed(True)
        else:
            self._db_manager.db.move_group(self._element, dest.group)
            dest.filter_changed(False)
            old_location.filter_changed(False)

        old_location.emit_children_changed(1, 0)
        dest.emit_children_changed(0, 1)

        old_location.updated()
        dest.updated()

    @property
    def element(self) -> Entry | Group:
        return self._element

    @GObject.Property(type=str, default="")
    def name(self) -> str:
        """Get element title or name.

        :returns: name or an empty string if there is none
        :rtype: str
        """
        if self._name is None:
            if self.is_group:
                self._name = self._element.name or ""
            else:
                self._name = self._element.title or ""

        return self._name

    @name.setter  # type: ignore
    def name(self, new_name: str) -> None:
        """Set entry title.

        :param str new_name: new title
        """
        if new_name != self._name:
            self._name = new_name
            if self.is_group:
                self._element.name = new_name
                self._db_manager.emit("sorting_changed", False)
            else:
                self._element.title = new_name
                self._db_manager.emit("sorting_changed", True)

            if self.is_entry:
                found, pos = self._db_manager.entries.find(self)
                if found:
                    self._db_manager.entries.items_changed(pos, 1, 1)
            else:
                found, pos = self._db_manager.groups.find(self)
                if found:
                    self._db_manager.groups.items_changed(pos, 1, 1)

            self.updated()

    @GObject.Property(type=str, default="")
    def notes(self) -> str:
        """Get entry notes.

        :returns: notes or an empty string if there is none
        :rtype: str
        """
        if self._notes is None:
            self._notes = self._element.notes or ""

        return self._notes

    @notes.setter  # type: ignore
    def notes(self, new_notes: str) -> None:
        """Set entry notes.

        :param str new_notes: new notes
        """
        if new_notes != self._notes:
            self._notes = new_notes
            self._element.notes = new_notes
            self.updated()

    @property
    def parentgroup(self) -> SafeGroup:
        """Parent Group of the element.

        :returns: parent group
        :rtype: SafeGroup
        """
        if self.is_root_group:
            return self

        for group in self._db_manager.groups:
            if group.uuid == self._element.parentgroup.uuid:
                return group

        logging.error("This should be unreachable: parentgroup")
        return SafeGroup(self._db_manager, self._element.parentgroup)

    @property
    def parentgroup_uuid(self) -> UUID:
        """UUID of the parent Group of the element.

        This method should be preferred than parentgroup since it does not go
        through the entire list of elements.

        :returns: parent group
        :rtype: SafeGroup
        """
        if self.is_root_group:
            return self.uuid

        return self._element.parentgroup.uuid

    @property
    def is_root_group(self) -> bool:
        if self.is_entry:
            return False

        return self._element.is_root_group

    @property
    def is_trash_bin(self) -> bool:
        if trash_bin_inner := self._db_manager.db.recyclebin_group:
            return self.uuid == trash_bin_inner.uuid

        return False

    @property
    def uuid(self) -> UUID:
        """UUID of the element.

        :returns: uuid of the element
        :rtype: UUID
        """
        return self._element.uuid

    @GObject.Property(type=str, flags=GObject.ParamFlags.READABLE)
    def parent_uuid_str(self) -> str:
        """UUID of the parent group of the element.

        :returns: uuid of the parent
        :rtype: UUID
        """
        if self.is_root_group:
            return ""

        return str(self.parentgroup_uuid)

    @property
    def path(self) -> list[str]:
        return self._element.path

    @property
    def atime(self) -> GLib.DateTime | None:
        """The UTC accessed time of the element."""
        try:
            time = self._element.atime
        except ValueError:
            logging.exception("Invalid accessed time")
            return None
        except OverflowError:
            logging.exception("Accessed time for %s overflows", self.name)
            return None
        except Exception:  # pylint: disable=broad-except
            logging.exception("Could not read accessed time")
            return None

        if not time:
            return None

        return GLib.DateTime.new_utc(
            time.year,
            time.month,
            time.day,
            time.hour,
            time.minute,
            time.second,
        )

    @GObject.Property(type=int, flags=GObject.ParamFlags.READABLE)
    def ctime_int(self) -> int:
        try:
            time = self._element.ctime
        except ValueError:
            logging.exception("Invalid creation time")
            return -1
        except OverflowError:
            logging.exception("Creation time for %s overflows", self.name)
            return -1
        except Exception:  # pylint: disable=broad-except
            logging.exception("Could not read creation time")
            return -1

        if not time:
            return -1

        return time.timestamp()

    @property
    def ctime(self) -> GLib.DateTime | None:
        """The UTC creation time of the element."""
        try:
            time = self._element.ctime
        except ValueError:
            logging.exception("Invalid creation time")
            return None
        except OverflowError:
            logging.exception("Creation time for %s overflows", self.name)
            return None
        except Exception:  # pylint: disable=broad-except
            logging.exception("Could not read creation time")
            return None

        if not time:
            return None

        return GLib.DateTime.new_utc(
            time.year,
            time.month,
            time.day,
            time.hour,
            time.minute,
            time.second,
        )

    @property
    def mtime(self) -> GLib.DateTime | None:
        """The UTC modified time of the element."""
        try:
            time = self._element.mtime
        except ValueError:
            logging.exception("Invalid modified time")
            return None
        except OverflowError:
            logging.exception("Modified time for %s overflows", self.name)
            return None
        except Exception:  # pylint: disable=broad-except
            logging.exception("Could not read modified time")
            return None

        if not time:
            return None

        return GLib.DateTime.new_utc(
            time.year,
            time.month,
            time.day,
            time.hour,
            time.minute,
            time.second,
        )


class SafeGroup(SafeElement):
    _entries = None
    _entries_filter = None
    _subgroups = None
    _subgroups_filter = None

    children_changed = GObject.Signal(arg_types=(int, int))

    def __init__(self, db_manager: DatabaseManager, group: Group) -> None:
        """GObject to handle a safe group.

        :param DatabaseManager db_manager:  database of the group
        :param Group group: group to handle
        """
        super().__init__(db_manager, group)

        if self.is_root_group:
            self._db_manager.root = self

        if self.is_trash_bin:
            self._db_manager.trash_bin = self

    @staticmethod
    def get_root(db_manager: DatabaseManager) -> SafeGroup:
        """Get the root group."""
        # pylint: disable=no-member
        if root := db_manager.root:
            return root

        for group in db_manager.groups:
            if group.is_root_group:
                return group

        logging.error("This should be unreachable: get_root")
        return SafeGroup(db_manager, db_manager.db.root_group)

    def new_entry(
        self,
        title: str = "",
        username: str = "",
        password: str = "",
    ) -> SafeEntry:
        """Add a new entry to self."""
        group = self.group
        force: bool = self._db_manager.check_entry_in_group_exists("", group)

        if username == "":
            username = self._db_manager.default_username

        new_entry = self._db_manager.db.add_entry(
            group,
            title,
            username,
            password,
            url=None,
            notes=None,
            expiry_time=None,
            tags=None,
            icon="0",
            force_creation=force,
        )
        safe_entry = SafeEntry(self._db_manager, new_entry)
        self.updated()
        self._db_manager.entries.append(safe_entry)

        return safe_entry

    def new_subgroup(
        self,
        name: str = "",
        icon: str | None = None,
        notes: str = "",
    ) -> SafeGroup:
        """Add a new subgroup to self."""
        new_group = self._db_manager.db.add_group(
            self.group,
            name,
            icon=icon,
            notes=notes,
        )
        safe_group = SafeGroup(self._db_manager, new_group)
        self.updated()
        self._db_manager.groups.append(safe_group)

        return safe_group

    def filter_changed(self, is_entry: bool) -> None:
        if is_entry:
            if self._entries_filter is None:
                self.init_entries()

            self._entries_filter.changed(Gtk.FilterChange.DIFFERENT)  # type: ignore
        else:
            if self._subgroups_filter is None:
                self.init_subgroups()

            self._subgroups_filter.changed(Gtk.FilterChange.DIFFERENT)  # type: ignore

    def init_subgroups(self):
        expr = Gtk.PropertyExpression.new(
            SafeElement.__gtype__, None, "parent_uuid_str"
        )
        self._subgroups_filter = Gtk.StringFilter.new(expr)
        self._subgroups_filter.props.search = str(self.uuid)
        self._subgroups = Gtk.FilterListModel.new(
            self._db_manager.groups,
            self._subgroups_filter,
        )

    def init_entries(self):
        expr = Gtk.PropertyExpression.new(
            SafeElement.__gtype__, None, "parent_uuid_str"
        )
        self._entries_filter = Gtk.StringFilter.new(expr)
        self._entries_filter.props.search = str(self.uuid)
        self._entries = Gtk.FilterListModel.new(
            self._db_manager.entries,
            self._entries_filter,
        )

    @property
    def subgroups(self) -> Gio.ListModel:
        if self._subgroups is None:
            self.init_subgroups()

        return self._subgroups

    @property
    def entries(self) -> Gio.ListModel:
        if self._entries is None:
            self.init_entries()

        return self._entries

    @property
    def group(self) -> Group:
        """Returns the private pykeepass group."""
        return self._element

    def emit_children_changed(self, removed, added):
        self.emit(self.children_changed, removed, added)


class SafeEntry(SafeElement):
    # pylint: disable=too-many-instance-attributes, too-many-public-methods

    _color_key = "color_prop_LcljUMJZ9X"
    _expired_id: int | None = None
    _note_key = "Notes"
    _otp: TOTP | None = None
    _otp_key = "otp"
    _attributes: AttributesModel | None = None
    _attachments: list[Attachment] | None = None
    _username: str | None = None
    _url: str | None = None
    _password: str | None = None
    _color: EntryColor | None = None
    _icon_nr: str | None = None

    history_saved = GObject.Signal()

    def __init__(self, db_manager: DatabaseManager, entry: Entry) -> None:
        """GObject to handle a safe entry.

        :param DatabaseManager db_manager:  database of the entry
        :param Entry entry: entry to handle
        """
        super().__init__(db_manager, entry)

        if otp_uri := entry.otp:
            try:
                if otp_uri.startswith("otpauth://"):
                    self._otp = parse_uri(otp_uri)  # type: ignore
                else:
                    self._otp = TOTP(otp_uri)
            except ValueError:
                logging.exception("Could not parse OTP")

        self._check_expiration()

    @property
    def entry(self) -> Entry:
        """Get entry.

        :returns: entry
        :rtype: Entry
        """
        return self._element

    def duplicate(self) -> None:
        """Duplicate an entry."""
        title: str = self.name or ""
        username: str = self.username or ""
        password: str = self.password or ""

        # NOTE: With clone is meant a duplicated object, not the process
        # of cloning/duplication; "the" clone
        entry = self.entry
        clone_entry: Entry = self._db_manager.db.add_entry(
            entry.parentgroup,
            title + " - " + _("Clone"),
            username,
            password,
            url=entry.url,
            notes=entry.notes,
            expiry_time=entry.expiry_time,
            tags=entry.tags,
            icon=entry.icon,
            force_creation=True,
        )
        clone_entry.expires = entry.expires
        if entry.otp:
            clone_entry.otp = entry.otp

        # Add custom properties
        for key in entry.custom_properties:
            value: str = entry.custom_properties[key] or ""
            clone_entry.set_custom_property(key, value)

        safe_entry = SafeEntry(self._db_manager, clone_entry)

        self.parentgroup.updated()
        self._db_manager.entries.append(safe_entry)

    def _check_expiration(self) -> None:
        """Check expiration.

        If the entry is expired, this ensures that a notification is sent.
        If the entry is not expired yet, a timeout is set to regularly
        check if the entry is expired.
        """
        if self._expired_id:
            GLib.source_remove(self._expired_id)
            self._expired_id = None

        if not self.props.expires:
            return

        # Update state
        self.notify("expired")

        if not self.props.expired:
            self._expired_id = GLib.timeout_add_seconds(600, self._is_expired)

    def _is_expired(self) -> bool:
        if self.props.expired:
            self._expired_id = None
            self.notify("expired")
            return GLib.SOURCE_REMOVE

        return GLib.SOURCE_CONTINUE

    def save_history(self) -> None:
        """Save current version of the entry in its history."""
        # NOTE Attachments are references, so duplicating them is ok.
        self._element.save_history()
        self.updated()
        self.emit(self.history_saved)

    def delete_history(self, entry: SafeEntry) -> None:
        """Delete entry from the history of self."""
        self._element.delete_history(entry.entry)
        self.updated()
        self.emit(self.history_saved)

    @GObject.Property(type=object, flags=GObject.ParamFlags.READABLE)
    def attachments(self) -> list[Attachment]:
        if self._attachments is None:
            self._attachments: list[Attachment] = self._element.attachments or []

        return self._attachments

    def add_attachment(self, byte_buffer: bytes, filename: str) -> Attachment:
        """Add an attachment to the entry.

        :param bytes byte_buffer: attachment content
        :param str filename: attachment name
        :returns: attachment
        :rtype: Attachment
        """
        assert self._attachments is not None

        attachment_id = self._db_manager.db.add_binary(byte_buffer)
        attachment = self._element.add_attachment(attachment_id, filename)
        self._attachments.append(attachment)
        self.updated()
        self.notify("attachments")

        return attachment

    def delete_attachment(self, attachment: Attachment) -> None:
        """Remove an attachment from the entry.

        :param Attachmennt attachment: attachment to delete
        """
        assert self._attachments is not None

        self._db_manager.db.delete_binary(attachment.id)
        self._attachments.remove(attachment)
        self.notify("attachments")
        self.updated()

    def get_attachment(self, id_: str) -> Attachment | None:
        """Get an attachment from its id.

        :param str id_: attachment id
        :returns: attachment
        :rtype: Attachment
        """
        assert self._attachments is not None

        for attachment in self._attachments:
            if str(attachment.id) == id_:
                return attachment

        return None

    def get_attachment_content(self, attachment: Attachment) -> bytes:
        """Get an attachment content.

        :param Attachmennt attachment: attachment
        """
        return self._db_manager.db.binaries[attachment.id]

    @GObject.Property(type=object, flags=GObject.ParamFlags.READABLE)
    def attributes(self) -> AttributesModel:
        if self._attributes is None:
            # NOTE Can fail at libpykeepass, see
            # https://github.com/libkeepass/pykeepass/issues/254
            # TODO Read as many attributes as possible until we see an error.
            try:
                attributes = {
                    key: value
                    for key, value in self._element.custom_properties.items()
                    if key not in (self._color_key, self._note_key, self._otp_key)
                }
            except Exception:  # pylint: disable=broad-except
                logging.exception("Could not read attributes")
                attributes = {}

            self._attributes = AttributesModel(attributes)

        return self._attributes

    def has_attribute(self, key: str) -> bool:
        """Check if an attribute exists.

        :param str key: attribute key to check
        """
        assert self._attributes is not None

        return self._attributes.has_attribute(key)

    def set_attribute(self, key: str, value: str, protected: bool = False) -> None:
        """Add or replace an entry attribute.

        :param str key: attribute key
        :param str value: attribute value
        """
        assert self._attributes is not None

        if self.props.attributes.get(key) == value:
            return

        if self._element.get_custom_property(
            key,
        ) == value and protected == self._element.is_custom_property_protected(key):
            return

        self._element.set_custom_property(key, value, protect=protected)

        self._attributes.insert(key, value)

        self.updated()
        self.notify("attributes")

    def delete_attribute(self, key: str) -> None:
        """Delete an attribute.

        :param key: attribute key to delete
        """
        assert self._attributes is not None

        if not self.has_attribute(key):
            return

        self._element.delete_custom_property(key)
        self._attributes.pop(key)
        self.updated()
        self.notify("attributes")

    def is_attribute_protected(self, key: str) -> bool:
        """Return whether the attribute with a specific key is protected.

        If there is no such key returns False.
        """
        return self.entry.is_custom_property_protected(key)

    @GObject.Property(type=str, default=EntryColor.NONE.value)
    def color(self) -> str:
        """Get entry color.

        :returns: color as string
        :rtype: str
        """
        if self._color is None:
            color_value: str = self._element.get_custom_property(self._color_key)
            self._color: str = color_value or EntryColor.NONE.value

        return self._color

    @color.setter  # type: ignore
    def color(self, new_color: str) -> None:
        """Set an entry color.

        :param str new_color: new color as string
        """
        if new_color != self._color:
            self._color = new_color
            self._element.set_custom_property(self._color_key, new_color)
            self.updated()

    @property
    def history(self) -> list[SafeEntry]:
        history = self._element.history
        return [SafeEntry(self._db_manager, entry) for entry in history]

    @GObject.Property(type=object)
    def icon(self) -> Icon:
        """Get icon number.

        :returns: icon number or "0" if no icon
        :rtype: str
        """
        if self._icon_nr is None:
            self._icon_nr: str = self._element.icon or ""

        try:
            return ICONS[self._icon_nr]
        except KeyError:
            return ICONS["0"]

    @icon.setter  # type: ignore
    def icon(self, new_icon_nr: str) -> None:
        """Set icon number.

        :param str new_icon_nr: new icon number
        """
        if new_icon_nr != self._icon_nr:
            self._icon_nr = new_icon_nr
            self._element.icon = new_icon_nr
            self.notify("icon-name")
            self.updated()

    @GObject.Property(type=str, default="", flags=GObject.ParamFlags.READABLE)
    def icon_name(self) -> str:
        """Get the icon name.

        :returns: icon name or the default icon if undefined
        :rtype: str
        """
        return self.props.icon.name

    @property
    def tags(self) -> dict:
        return self.entry.tags

    @tags.setter  # type: ignore
    def tags(self, new_tags: dict) -> None:
        self.entry.tags = new_tags
        self._db_manager.is_dirty = True

    @GObject.Property(type=str, default="")
    def otp(self) -> str:
        if self._otp:
            return self._otp.secret

        return ""

    @otp.setter  # type: ignore
    def otp(self, otp: str) -> None:
        updated = False

        # Some sites give the secret in chunks split by spaces for easy reading
        # lets strip those as they'll produce an invalid secret.
        otp = otp.replace(" ", "")

        if not otp and self._otp:
            # Delete existing
            self._otp = None
            # NOTE the opt property doesn't accept None.
            self._element.otp = ""
            self.updated()
        elif self._otp and self._otp.secret != otp:
            # Changing an existing OTP
            self._otp.secret = otp
            updated = True
        elif otp:
            # Creating brand new OTP.
            self._otp = TOTP(otp, issuer=self.name)
            updated = True

        if updated:
            self._element.otp = self._otp.provisioning_uri()
            self.updated()

    def otp_interval(self) -> int:
        if isinstance(self._otp, TOTP):
            return self._otp.interval

        return 30

    def otp_lifespan(self) -> float | None:
        """Return seconds until token expires."""
        if isinstance(self._otp, TOTP):
            gnow = GLib.DateTime.new_now_utc()
            now_seconds = gnow.to_unix()
            now_milis = gnow.get_seconds() % 1
            now = now_seconds + now_milis
            return self._otp.interval - now % self._otp.interval

        return None

    def otp_token(self) -> str | None:
        if self._otp:
            try:
                return self._otp.now()
            except binascii.Error:
                logging.debug(
                    "Error caught in OTP token generation (likely invalid "
                    "base32 secret).",
                )

        return None

    @GObject.Property(type=str, default="")
    def password(self) -> str:
        """Get entry password.

        :returns: password or an empty string if there is none
        :rtype: str
        """
        if self._password is None:
            self._password: str = self._element.password or ""

        return self._password

    @password.setter  # type: ignore
    def password(self, new_password: str) -> None:
        """Set entry password.

        :param str new_password: new password
        """
        if new_password != self._password:
            self._password = new_password
            self._element.password = new_password
            self.updated()

    @GObject.Property(type=str, default="")
    def url(self) -> str:
        """Get entry url.

        :returns: url or an empty string if there is none
        :rtype: str
        """
        if self._url is None:
            self._url: str = self._element.url or ""

        return self._url

    @url.setter  # type: ignore
    def url(self, new_url: str) -> None:
        """Set entry url.

        :param str new_url: new url
        """
        if new_url != self._url:
            self._url = new_url
            self._element.url = new_url
            self.updated()

    @GObject.Property(type=str, default="")
    def username(self) -> str:
        """Get entry username.

        :returns: username or an empty string if there is none
        :rtype: str
        """
        if self._username is None:
            self._username: str = self._element.username or ""

        return self._username

    @username.setter  # type: ignore
    def username(self, new_username: str) -> None:
        """Set entry username.

        :param str new_username: new username
        """
        if new_username != self._username:
            self._username = new_username
            self._element.username = new_username
            self.updated()

    @GObject.Property(type=bool, default=False)
    def expires(self) -> bool:
        return self.entry.expires

    @expires.setter  # type: ignore
    def expires(self, value: bool) -> None:
        if value != self.entry.expires:
            self.entry.expires = value
            self._check_expiration()
            self.updated()
            self.notify("expired")

    @GObject.Property(
        type=bool,
        default=False,
        flags=GObject.ParamFlags.READABLE | GObject.ParamFlags.EXPLICIT_NOTIFY,
    )
    def expired(self):
        try:
            return self.entry.expired
        except Exception:  # pylint: disable=broad-except
            logging.exception("Could not read expiry date from %s", self.name)
            return False

    @property
    def expiry_time(self) -> GLib.DateTime | None:
        """Returns the expiration time in the UTC timezone.

        Returns None when there isn't an expiration date.
        """
        try:
            time = self.entry.expiry_time
        except Exception:  # pylint: disable=broad-except
            logging.exception("Could not read expiry date from %s", self.name)
            return None

        if not time:
            return None

        return GLib.DateTime.new_utc(
            time.year,
            time.month,
            time.day,
            time.hour,
            time.minute,
            time.second,
        )

    @expiry_time.setter  # type: ignore
    def expiry_time(self, value: GLib.DateTime) -> None:
        """Set the expiration time in the UTC timezone."""
        if value != self.entry.expiry_time:
            expired = datetime(
                value.get_year(),
                value.get_month(),
                value.get_day_of_month(),
                value.get_hour(),
                value.get_minute(),
                value.get_second(),
                tzinfo=UTC,
            )
            self.entry.expiry_time = expired
            self._check_expiration()

            self.updated()


class Icon(NamedTuple):
    # pylint: disable=inherit-non-class
    # This is a false positive because pylint does not properly handle
    # Python 3.9 at the moment. It can be safely removed once pylint
    # handles it.
    name: str
    visible: bool = False


# https://github.com/dlech/KeePass2.x/blob/4facf2f1ebc76eeddbe11975eccb0dc2b49dfc37/KeePassLib/PwEnums.cs#L81
# https://hsto.org/files/b1e/d20/e38/b1ed20e385d642cc870355fdef153fb9.png
# FIXME: Based on the names from the links above, some of the current
# icons should be replaced.
ICONS = {
    "0": Icon("dialog-password-symbolic", True),
    "1": Icon("network-wired-symbolic", True),
    "2": Icon("dialog-warning-symbolic"),
    "3": Icon("network-server-symbolic"),
    "4": Icon("document-edit-symbolic"),
    "5": Icon("media-view-subtitles-symbolic"),
    "6": Icon("application-x-addon-symbolic"),
    "7": Icon("notepad-symbolic"),
    "8": Icon("network-wired-symbolic"),
    "9": Icon("send-symbolic", True),
    "10": Icon("text-x-generic-symbolic"),
    "11": Icon("camera-photo-symbolic"),
    "12": Icon("network-wireless-signal-excellent-symbolic", True),
    "13": Icon("dialog-password-symbolic"),
    "14": Icon("colorimeter-colorhug-symbolic"),
    "15": Icon("scanner-symbolic"),
    "16": Icon("user-available-symbolic", True),
    "17": Icon("media-optical-cd-audio-symbolic"),
    "18": Icon("video-display-symbolic"),
    "19": Icon("mail-unread-symbolic", True),
    "20": Icon("settings-symbolic"),
    "21": Icon("edit-paste-symbolic"),
    "22": Icon("edit-paste-symbolic"),
    "23": Icon("display-with-window-symbolic", True),
    "24": Icon("uninterruptible-power-supply-symbolic"),
    "25": Icon("mail-unread-symbolic"),
    "26": Icon("media-floppy-symbolic"),
    "27": Icon("drive-harddisk-symbolic", True),
    "28": Icon("dialog-password-symbolic"),
    "29": Icon("airplane-mode-symbolic", True),
    "30": Icon("terminal-symbolic", True),
    "31": Icon("printer-symbolic"),
    "32": Icon("image-x-generic-symbolic"),
    "33": Icon("edit-select-all-symbolic"),
    "34": Icon("preferences-system-symbolic", True),
    "35": Icon("network-workgroup-symbolic"),
    "36": Icon("dialog-password-symbolic"),
    "37": Icon("auth-fingerprint-symbolic", True),
    "38": Icon("drive-harddisk-symbolic"),
    "39": Icon("document-open-recent-symbolic"),
    "40": Icon("system-search-symbolic"),
    "41": Icon("applications-games-symbolic", True),
    "42": Icon("media-flash-symbolic"),
    "43": Icon("user-trash-symbolic"),
    "44": Icon("notepad-symbolic"),
    "45": Icon("edit-delete-symbolic"),
    "46": Icon("dialog-question-symbolic"),
    "47": Icon("package-x-generic-symbolic"),
    "48": Icon("folder-symbolic", True),
    "49": Icon("folder-open-symbolic"),
    "50": Icon("document-open-symbolic"),
    "51": Icon("system-lock-screen-symbolic", True),
    "52": Icon("rotation-locked-symbolic"),
    "53": Icon("object-select-symbolic"),
    "54": Icon("document-edit-symbolic"),
    "55": Icon("image-x-generic-symbolic"),
    "56": Icon("open-book-symbolic", True),
    "57": Icon("view-list-symbolic"),
    "58": Icon("avatar-default-symbolic", True),
    "59": Icon("applications-engineering-symbolic"),
    "60": Icon("go-home-symbolic"),
    "61": Icon("starred-symbolic", True),
    "62": Icon("start-here-symbolic"),
    "63": Icon("dialog-password-symbolic"),
    "64": Icon("start-here-symbolic"),
    "65": Icon("open-book-symbolic"),
    "66": Icon("money-symbolic", True),
    "67": Icon("application-certificate-symbolic"),
    "68": Icon("phone-apple-iphone-symbolic", True),
}



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gi.repository import Adw, Gio, GLib, Gtk


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/references_dialog.ui")
class ReferencesDialog(Adw.Dialog):
    __gtype_name__ = "ReferencesDialog"

    __property = "P"

    _codes_popover = Gtk.Template.Child()
    _property_label = Gtk.Template.Child()
    _property_popover_button = Gtk.Template.Child()
    _reference_entry = Gtk.Template.Child()
    _uuid_label = Gtk.Template.Child()
    _uuid_popover = Gtk.Template.Child()

    def __init__(self, unlocked_database):
        super().__init__()

        self.__unlocked_database = unlocked_database
        self.__database_manager = unlocked_database.database_manager
        self.__signal_handle = 0

        self.__setup_actions()
        self.__setup_signals()
        self.__setup_widgets()

    def __setup_widgets(self) -> None:
        self.__update_reference_entry()

    def __setup_signals(self) -> None:
        self.__connect_model_buttons_signals()
        handle = self.__database_manager.connect("notify::locked", self.__on_locked)
        self.__signal_handle = handle

    def __setup_actions(self) -> None:
        action_group = Gio.SimpleActionGroup.new()
        for label in ["U", "N", "A", "P", "T"]:
            simple_action = Gio.SimpleAction.new(label, None)
            simple_action.connect("activate", self.__on_property_model_button_clicked)
            action_group.insert(simple_action)
        self.insert_action_group("reference", action_group)

    def __connect_model_buttons_signals(self):
        self._uuid_popover.set_parent(self._uuid_label)
        self._codes_popover.set_parent(self._property_label)

    def __update_reference_entry(self):
        """Update the reference entry and selected label text."""
        uuid = self.__unlocked_database.current_element.uuid
        encoded_uuid = uuid.hex.upper()

        self._property_popover_button.set_label(self.__property)

        self._reference_entry.set_text(
            "{REF:" + self.__property + "@I:" + encoded_uuid + "}",
        )

    def do_closed(self) -> None:
        if handle := self.__signal_handle:
            self.__database_manager.disconnect(handle)
            self.__signal_handle = 0

    @Gtk.Template.Callback()
    def _open_codes_popover(self, gesture, _n_points, _x, _y):
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        self._codes_popover.popup()

    @Gtk.Template.Callback()
    def _open_uuid_popover(self, gesture, _n_points, _x, _y):
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        self._uuid_popover.popup()

    @Gtk.Template.Callback()
    def _on_copy_secondary_button_clicked(self, entry, _position):
        self.__unlocked_database.start_database_lock_timer()
        self.__unlocked_database.clipboard.set(entry.get_text())

    def __on_property_model_button_clicked(
        self,
        action: Gio.Action,
        _parameter: GLib.Variant,
    ) -> None:
        self.__unlocked_database.start_database_lock_timer()
        self.__property = action.props.name
        self.__update_reference_entry()

    def __on_locked(self, database_manager, _value):
        locked = database_manager.props.locked
        if locked:
            self.close()



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
import secrets
import stat
import typing
from gettext import gettext as _
from pathlib import Path

from Cryptodome.Cipher import AES
from Cryptodome.Random import get_random_bytes
from gi.repository import Gio, GLib, Gtk

if typing.TYPE_CHECKING:
    from collections.abc import Callable


ATTRIBUTE_HOST_PATH = "xattr::document-portal.host-path"


def format_time(time: GLib.DateTime | None, hours: bool = True) -> str:
    """Display a UTC DateTime in the local timezone."""
    if not time:
        return ""

    time_format = "%e %b %Y"
    if hours:
        time_format += " %R"  # NOTE This is a U+2002 En Space.

    return time.to_local().format(time_format)


def create_random_data(bytes_buffer):
    return secrets.token_bytes(bytes_buffer)


def generate_keyfile_sync(gfile: Gio.File) -> str:
    """Generate a keyfile.

    The callback returns a GFile as its source object and the keyfile hash as
    its user_data. Returns the hash of the keyfile.
    """
    key = get_random_bytes(32)
    cipher = AES.new(key, AES.MODE_EAX)
    ciphertext, tag = cipher.encrypt_and_digest(create_random_data(96))  # type: ignore
    contents = cipher.nonce + tag + ciphertext  # type: ignore
    keyfile_hash = GLib.compute_checksum_for_data(GLib.ChecksumType.SHA1, contents)

    flags = Gio.FileCreateFlags.REPLACE_DESTINATION | Gio.FileCreateFlags.PRIVATE
    gfile.replace_contents(contents, None, False, flags, None)
    # Sets the file as read-only
    if path := gfile.get_path():
        Path(path).chmod(stat.S_IREAD)

    return keyfile_hash


def generate_keyfile_async(gfile: Gio.File, callback: Gio.AsyncReadyCallback) -> None:
    def generate_keyfile_task(task, obj, _data, _cancellable):
        try:
            keyfile_hash = generate_keyfile_sync(obj)
        except GLib.Error as err:  # pylint: disable=broad-except
            task.return_error(err)
        else:
            task.return_value(keyfile_hash)

    task = Gio.Task.new(gfile, None, callback)
    task.run_in_thread(generate_keyfile_task)


def generate_keyfile_finish(result: Gio.AsyncResult) -> tuple[bool, str]:
    return result.propagate_value()


def compare_passwords(pass1: str | None, pass2: str | None) -> bool:
    if pass1 is not None and pass2 is not None:
        return secrets.compare_digest(bytes(pass1, "utf-8"), bytes(pass2, "utf-8"))

    return pass1 is None and pass2 is None


async def get_host_path(gfile: Gio.File) -> str | None:
    """Tries to get the file host-path, if not set will return the file path."""
    try:
        info = await gfile.query_info_async(
            ATTRIBUTE_HOST_PATH, Gio.FileQueryInfoFlags.NONE, GLib.PRIORITY_DEFAULT
        )
    except GLib.Error:
        logging.exception("Could not query file info")
        return gfile.get_path()

    try:
        host_path = info.get_attribute_string(ATTRIBUTE_HOST_PATH)
    except GLib.Error:
        logging.exception("Could not query file attribute")
    else:
        return host_path or gfile.get_path()

    return gfile.get_path()


class KeyFileFilter:
    """Filter out Keyfiles in the file chooser dialog."""

    def __init__(self):
        self.file_filter = Gtk.FileFilter()

        self.file_filter.set_name(_("Keyfile"))
        self.file_filter.add_mime_type("application/octet-stream")
        self.file_filter.add_mime_type("application/x-keepass2")
        self.file_filter.add_mime_type("text/plain")
        self.file_filter.add_mime_type("application/x-iwork-keynote-sffkey")


class LazyValue[T]:
    """A lazy value, i.e. a value which is only computed when it is actually needed."""

    def __init__(self, compute: Callable[[], T]):
        self._value: T | None = None
        self._compute = compute

    @property
    def value(self) -> T:
        """Get the value (and compute it if not done yet)."""
        if self._value is None:
            self._value = self._compute()
        return self._value



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
from gettext import gettext as _

from gi.repository import Adw, GLib, Gtk


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/save_dialog.ui")
class SaveDialog(Adw.AlertDialog):
    __gtype_name__ = "SaveDialog"

    def __init__(self, window):
        super().__init__()

        self.window = window

    @Gtk.Template.Callback()
    def _on_discard(self, _dialog, _response):
        self.window.force_close()

    @Gtk.Template.Callback()
    def _on_save(self, _dialog, _response):
        if database := self.window.unlocked_db:
            database.database_manager.save_async(self._save_finished)

    def _save_finished(self, database_manager, result):
        try:
            is_saved = database_manager.save_finish(result)
        except GLib.Error:
            self.window.send_notification(_("Could not save Safe"))
            logging.exception("Could not save Safe")
        else:
            if is_saved:
                self.window.close()
            else:
                # This shouldn't happen
                self.window.send_notification(_("Could not save Safe"))



# SPDX-License-Identifier: GPL-3.0-only
from gi.repository import Adw, Gtk


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/entry_page_tag.ui")
class EntryPageTag(Adw.Bin):
    __gtype_name__ = "EntryPageTag"

    _label = Gtk.Template.Child()
    remove_button = Gtk.Template.Child()

    def __init__(self, name):
        super().__init__()

        self._label.set_label(name)



# SPDX-License-Identifier: GPL-3.0-only
"""GUI Page and function in order to create a new Safe."""

from __future__ import annotations

import logging
from gettext import gettext as _

from gi.repository import Adw, GLib, Gtk

from gsecrets.password_generator_popover import PasswordGeneratorPopover


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/create_database.ui")
class CreateDatabase(Adw.Bin):
    """Creates a new Safe when invoked."""

    __gtype_name__ = "CreateDatabase"

    stack = Gtk.Template.Child()

    password_row = Gtk.Template.Child()
    password_confirm_row = Gtk.Template.Child()
    provider_group = Gtk.Template.Child()
    create_button = Gtk.Template.Child()
    open_safe_button = Gtk.Template.Child()
    back_button = Gtk.Template.Child()
    match_hint = Gtk.Template.Child()
    banner = Gtk.Template.Child()

    def __init__(self, window, dbm):
        super().__init__()

        self.database_manager = dbm
        self.window = window
        generator = Gtk.MenuButton()
        generator.set_icon_name("dice3-symbolic")
        generator_popover = PasswordGeneratorPopover()
        generator.set_popover(generator_popover)
        generator.set_tooltip_text(_("Generate New Password"))
        generator.add_css_class("flat")
        generator.set_valign(Gtk.Align.CENTER)
        generator_popover.connect("generated", self._on_password_generated)
        self.password_row.add_suffix(generator)
        self.password_row.connect("changed", self._on_password_changed)
        self.password_confirm_row.connect("changed", self._on_password_changed)

        self.back_button.props.sensitive = True
        self.back_button.connect("clicked", self.on_headerbar_back_button_clicked)
        for key_provider in self.window.key_providers.get_key_providers():
            if key_provider.available:
                self.provider_group.add(key_provider.create_database_row())

        self.window.set_default_widget(self.create_button)

    def do_realize(self):  # pylint: disable=arguments-differ
        Gtk.Widget.do_realize(self)
        self.password_row.grab_focus()

    def _on_password_changed(self, _entry: Gtk.Entry) -> None:
        is_match = self.password_row.props.text == self.password_confirm_row.props.text

        if not is_match:
            self.password_confirm_row.add_css_class("error")
            self.match_hint.set_label(_("Passwords do not match"))
        else:
            self.password_confirm_row.remove_css_class("error")
            self.match_hint.set_label("")

        self.create_button.set_sensitive(is_match and self.password_row.props.text)

    def success_page(self):
        self.stack.set_visible_child_name("safe-successfully-create")
        self.clear_input_fields()
        self.open_safe_button.grab_focus()

    def failure_page(self):
        self.stack.set_visible_child_name("select-auth-method")
        self.clear_input_fields()
        self.window.send_notification(_("Unable to create database"))

    def on_headerbar_back_button_clicked(self, _widget):
        self.go_back()

    def go_back(self):
        self.window.invoke_initial_screen()

    def _on_set_credentials(self, dbm, result):
        try:
            is_saved = dbm.set_credentials_finish(result)
        except GLib.Error:
            logging.exception("Could not set credentials")
            self.failure_page()
        else:
            if is_saved:
                self.success_page()
            else:  # Unreachable
                self.failure_page()

    def _on_password_generated(self, _popover, password):
        self.password_row.props.text = password
        self.password_confirm_row.props.text = password

    def _on_generate_composite_key(self, providers, result):
        self.stack.set_sensitive(True)

        try:
            composition_key = providers.generate_composite_key_finish(result)
        except GLib.Error:
            logging.exception("Failed to generate composite key")
            self.window.send_notification(_("Failed to generate composite key"))
            return

        self.database_manager.set_credentials_async(
            password=self.password_confirm_row.props.text,
            composition_key=composition_key,
            callback=self._on_set_credentials,
        )

    @Gtk.Template.Callback()
    def _on_create_button_clicked(self, _button: Gtk.Button) -> None:
        self.stack.set_sensitive(False)
        self.window.key_providers.generate_composite_key_async(
            self.database_manager.get_salt_as_lazy(),
            self._on_generate_composite_key,
        )

    @Gtk.Template.Callback()
    def on_finish_button_clicked(self, _widget: Gtk.Button) -> None:
        self.window.start_database_opening_routine(self.database_manager.path)

    def clear_input_fields(self) -> None:
        """Empty all Entry textfields."""
        self.password_row.set_text("")
        self.password_confirm_row.set_text("")

        for key_provider in self.window.key_providers.get_key_providers():
            key_provider.clear_input_fields()

    def show_banner(self, label: str) -> None:
        self.banner.set_title(label)
        self.banner.set_revealed(True)

    def close_banner(self):
        self.banner.set_revealed(False)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
import threading
import typing
from gettext import gettext as _
from pathlib import Path, PurePath

import pyhibp
import requests
from gi.repository import Adw, Gio, GLib, GObject, Gtk
from pyhibp import pwnedpasswords as pw

from gsecrets.password_generator import strength
from gsecrets.utils import (
    compare_passwords,
    format_time,
    get_host_path,
)

if typing.TYPE_CHECKING:
    from gsecrets.provider.base_provider import BaseProvider


WEAK = 0.5
MEDIUM = 1.5


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/database_settings_dialog.ui")
class DatabaseSettingsDialog(Adw.PreferencesDialog):
    # pylint: disable=too-many-instance-attributes

    __gtype_name__ = "DatabaseSettingsDialog"

    new_password: str | None = None
    current_keyfile_hash = None
    current_keyfile_path = None
    new_keyfile_hash = None
    new_keyfile_path = None

    entries_number = None
    groups_number = None
    passwords_number = None

    auth_apply_button = Gtk.Template.Child()

    level_bar = Gtk.Template.Child()

    encryption_algorithm_row = Gtk.Template.Child()
    date_row = Gtk.Template.Child()
    derivation_algorithm_row = Gtk.Template.Child()
    n_entries_row = Gtk.Template.Child()
    n_groups_row = Gtk.Template.Child()
    n_passwords_row = Gtk.Template.Child()
    name_row = Gtk.Template.Child()
    description_row = Gtk.Template.Child()
    default_username_row = Gtk.Template.Child()
    path_row = Gtk.Template.Child()
    size_row = Gtk.Template.Child()
    version_row = Gtk.Template.Child()
    current_password_entry = Gtk.Template.Child()
    provider_group = Gtk.Template.Child()
    banner = Gtk.Template.Child()
    unique_passwords_row = Gtk.Template.Child()
    reused_passwords_row = Gtk.Template.Child()
    weak_password_group = Gtk.Template.Child()
    hibp_group = Gtk.Template.Child()
    hibp_stack = Gtk.Template.Child()
    hibp_spinner = Gtk.Template.Child()
    hibp_refresh_button = Gtk.Template.Child()

    confirm_password_entry = Gtk.Template.Child()
    new_password_entry = Gtk.Template.Child()

    def __init__(self, unlocked_database):
        super().__init__()

        self.unlocked_database = unlocked_database
        self.database_manager = unlocked_database.database_manager
        self.window = self.unlocked_database.window
        self.signals = []
        self.bindings = []

        self.__setup_widgets()
        self.__setup_signals()

        self._providers = self.window.key_providers

        for key_provider in self._providers.get_key_providers():
            if key_provider.available:
                self.provider_group.add(key_provider.create_database_row())

                show_id = key_provider.connect(
                    key_provider.show_message,
                    self._on_show_message,
                )
                hide_id = key_provider.connect(
                    key_provider.hide_message,
                    self._on_hide_message,
                )

                self.signals.append((show_id, key_provider))
                self.signals.append((hide_id, key_provider))

        self._fill_audit()

    def _fill_audit(self):
        passwords = []
        weak_passwords_found = False

        for entry in self.database_manager.db.find_entries():
            if entry.password:
                passwords.append(entry.password)

                pwd_strength = strength(entry.password)
                if pwd_strength < MEDIUM:
                    weak_passwords_found = True

                    row = Adw.ActionRow()
                    if entry.title:
                        row.set_title(entry.title)
                    if entry.username:
                        row.set_subtitle(entry.username)

                    score = Gtk.Label()
                    if pwd_strength < WEAK:
                        score.set_label(_("Insecure"))
                    else:
                        score.set_label(_("Weak"))

                    row.add_suffix(score)
                    self.weak_password_group.add(row)

        if not weak_passwords_found:
            row = Adw.ActionRow()
            row.set_title(_("No weak passwords found"))
            self.weak_password_group.add(row)

        # Statistics
        reused_dict = {
            pwd: passwords.count(pwd) for pwd in passwords if passwords.count(pwd) > 1
        }
        self.reused_passwords_row.set_subtitle(str(len(reused_dict)))

        unique_dict = {
            pwd: passwords.count(pwd) for pwd in passwords if passwords.count(pwd) == 1
        }
        self.unique_passwords_row.set_subtitle(str(len(unique_dict)))

        thread = threading.Thread(target=self.check_network_access)
        thread.daemon = True
        thread.start()

        self.hibp_stack.set_visible_child_name("button")

    def check_network_access(self):
        reachable = True
        status_code_ok = True

        try:
            check_url = requests.get("https://haveibeenpwned.com/", timeout=3)
        except requests.exceptions.RequestException:
            reachable = False
        else:
            status_code_ok = check_url.status_code == requests.codes.ok

        GLib.idle_add(self.update_hibp_ui, reachable, status_code_ok)

    def update_hibp_ui(self, reachable, status_code_ok):
        if not reachable:
            self.hibp_refresh_button.set_sensitive(False)

            row = Adw.ActionRow()
            row.set_title("No Network Access Configured")
            row.set_subtitle("Add --share=network to your Flatpak manifest")
            self.hibp_group.add(row)
            return

        self.hibp_refresh_button.set_sensitive(status_code_ok)

    @Gtk.Template.Callback()
    def on_hibp_refresh_clicked(self, _entry: Gtk.Entry) -> None:
        self.hibp_stack.set_visible_child_name("spinner")
        self._hibp_refresh_async(self._on_hibp_refresh)

    def _hibp_refresh_async(self, callback: Gio.AsyncReadyCallback) -> None:
        def lookup(task, _source_object, _task_data, _cancellable):
            rows = []

            for entry in self.database_manager.db.find_entries():
                pyhibp.set_user_agent(ua="Secrets")
                if entry.password:
                    resp = pw.is_password_breached(password=entry.password)
                    if resp:
                        row = Adw.ActionRow()
                        if title := entry.title:
                            row.set_title(title)

                        if username := entry.username:
                            row.set_subtitle(username)

                        rows.append(row)

            task.return_value(rows)

        task = Gio.Task.new(self, None, callback)
        task.run_in_thread(lookup)

    def _hibp_refresh_finish(self, result):
        return result.propagate_value()

    def _on_hibp_refresh(
        self,
        _dialog: DatabaseSettingsDialog,
        result: Gio.AsyncResult,
    ) -> None:
        self.hibp_stack.set_visible(False)
        _success, rows = self._hibp_refresh_finish(result)

        if rows:
            for row in rows:
                self.hibp_group.add(row)
        else:
            row = Adw.ActionRow()
            row.set_title(_("No records found"))
            self.hibp_group.add(row)

    def do_closed(self):
        for signal_id, obj in self.signals:
            obj.disconnect(signal_id)

        self.signals = []

        for binding in self.bindings:
            binding.unbind()

        self.bindings = []

    def _on_show_message(self, _provider: BaseProvider, label: str) -> None:
        self.banner.set_title(label)
        self.banner.set_revealed(True)

    def _on_hide_message(self, _provider: BaseProvider) -> None:
        self.banner.set_revealed(False)

    #
    # Dialog Creation
    #

    def __setup_signals(self) -> None:
        signal_id = self.database_manager.connect("notify::locked", self.__on_locked)
        name_binding = self.database_manager.bind_property(
            "name",
            self.name_row,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        description_binding = self.database_manager.bind_property(
            "description",
            self.description_row,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

        self.signals.append((signal_id, self.database_manager))

        self.bindings.append(name_binding)
        self.bindings.append(description_binding)

    def __setup_widgets(self) -> None:
        # Dialog
        self.set_detail_values()

        stats_thread = threading.Thread(target=self.start_stats_thread)
        stats_thread.daemon = True
        stats_thread.start()

    @Gtk.Template.Callback()
    def on_password_entry_changed(self, _entry: Gtk.Entry) -> None:
        """CB if password entry (existing or new) has changed."""
        self.unlocked_database.start_database_lock_timer()

        new_password = self.new_password_entry.get_text()
        conf_password = self.confirm_password_entry.get_text()

        if new_password != conf_password:
            self.new_password_entry.add_css_class("error")
            self.confirm_password_entry.add_css_class("error")
        else:
            self.new_password_entry.remove_css_class("error")
            self.confirm_password_entry.remove_css_class("error")

        correct_input = self.passwords_coincide() and self.correct_credentials()
        self.auth_apply_button.set_sensitive(correct_input)

    def correct_credentials(self) -> bool:
        database_password = self.database_manager.password
        current_password = self.current_password_entry.get_text()
        return compare_passwords(database_password, current_password)

    def passwords_coincide(self) -> bool:
        new_password = self.new_password_entry.get_text()
        repeat_password = self.confirm_password_entry.get_text()

        return compare_passwords(new_password, repeat_password)

    def _on_set_credentials(self, database_manager, result):
        try:
            is_saved = database_manager.set_credentials_finish(result)
        except GLib.Error:
            logging.exception("Could not set credentials")
            self.add_toast(Adw.Toast.new(_("Could not apply changes")))
        else:
            if not is_saved:  # Should be unreachable
                logging.error("Credentials set without changes")
                self.add_toast(Adw.Toast.new(_("Could not apply changes")))
                return  # Still executes the finally block

            self.add_toast(Adw.Toast.new(_("Changes Applied")))

            # Restore all widgets
            self.current_password_entry.set_text("")
            self.new_password_entry.set_text("")
            self.confirm_password_entry.set_text("")

            self.current_keyfile_hash = None
            self.current_keyfile_path = None
            self.new_keyfile_hash = None
            self.new_keyfile_path = None
        finally:
            self.set_sensitive(True)
            self.auth_apply_button.set_sensitive(False)
            self.auth_apply_button.set_label(_("_Apply Changes"))
            for provider in self.provider_group:
                provider.set_sensitive(True)

    def _on_generate_composite_key(self, providers, result):
        for provider in self.provider_group:
            provider.set_sensitive(False)

        try:
            self._new_composition_key = providers.generate_composite_key_finish(result)
        except GLib.Error:
            logging.exception("Failed to generate composite key")
            self.window.send_notification(_("Failed to generate composite key"))
            return

        self.database_manager.check_file_changes_async(self.on_check_file_changes)

    @Gtk.Template.Callback()
    def on_default_username_changed(self, entry: Adw.EntryRow) -> None:
        self.database_manager.default_username = entry.get_text()

    @Gtk.Template.Callback()
    def on_auth_apply_button_clicked(self, button):
        # Insensitive entries and buttons
        self.set_sensitive(False)

        spinner = Adw.Spinner()
        button.set_child(spinner)
        button.set_sensitive(False)

        self.window.key_providers.generate_composite_key_async(
            self.database_manager.get_salt_as_lazy(),
            self._on_generate_composite_key,
        )

    def on_check_file_changes(self, dbm, result):
        try:
            conflicts = dbm.check_file_changes_finish(result)
        except GLib.Error:
            logging.exception("Could not monitor file changes")
            toast = _("Could not change credentials")
            self.add_toast(toast)
        else:
            if conflicts:
                dialog = Adw.AlertDialog.new(
                    _("Conflicts While Saving"),
                    _(
                        "The safe was modified from somewhere else. Please resolve these conflicts from the main window when saving.",  # noqa: E501
                    ),
                )
                dialog.add_response("ok", _("_OK"))
                dialog.present(self)
            else:
                new_password = self.new_password_entry.props.text
                self.database_manager.set_credentials_async(
                    new_password,
                    self._new_composition_key,
                    self._on_set_credentials,
                )
        finally:
            self.set_sensitive(True)
            self.auth_apply_button.set_sensitive(False)
            self.auth_apply_button.set_label(_("_Apply Changes"))
            self._new_composition_key = None

    @Gtk.Template.Callback()
    def on_password_generated(self, _popover, password):
        self.confirm_password_entry.props.text = password
        self.new_password_entry.props.text = password

    def set_detail_values(self):
        # Name
        self.default_username_row.props.text = self.database_manager.default_username

        # Path
        path = self.database_manager.path
        gfile = Gio.File.new_for_path(path)

        async def set_path():
            path = PurePath(await get_host_path(gfile))

            home = str(Path.home())
            if path.is_relative_to(home):
                self.path_row.props.subtitle = "~/" + str(path.relative_to(home))
            else:
                self.path_row.props.subtitle = str(path)

        app = Gio.Application.get_default()
        app.create_asyncio_task(set_path())

        # Size
        def query_info_cb(gfile, result):
            try:
                file_info = gfile.query_info_finish(result)
            except GLib.Error:
                logging.exception("Could not query file info")
            else:
                size = file_info.get_size()  # In bytes.
                self.size_row.props.subtitle = GLib.format_size(size)

        attributes = Gio.FILE_ATTRIBUTE_STANDARD_SIZE
        gfile.query_info_async(
            attributes,
            Gio.FileQueryInfoFlags.NONE,
            GLib.PRIORITY_DEFAULT,
            None,
            query_info_cb,
        )

        # Version
        version = self.database_manager.db.version
        self.version_row.props.subtitle = str(version[0]) + "." + str(version[1])

        # Date
        # TODO g_file_info_get_creation_date_time introduced in GLib 2.70.
        epoch_time = Path(path).stat().st_ctime  # Time since UNIX epoch.
        gdate = GLib.DateTime.new_from_unix_utc(epoch_time)
        self.date_row.props.subtitle = format_time(gdate)

        # Encryption Algorithm
        enc_alg = _("Unknown")
        enc_alg_priv = self.database_manager.db.encryption_algorithm
        if enc_alg_priv == "aes256":
            # NOTE: AES is a proper name
            enc_alg = _("AES 256-bit")
        elif enc_alg_priv == "chacha20":
            # NOTE: ChaCha20 is a proper name
            enc_alg = _("ChaCha20 256-bit")
        elif enc_alg_priv == "twofish":
            # NOTE: Twofish is a proper name
            enc_alg = _("Twofish 256-bit")
        self.encryption_algorithm_row.props.subtitle = enc_alg

        # Derivation Algorithm
        der_alg = _("Unknown")
        der_alg_priv = self.database_manager.db.kdf_algorithm
        if der_alg_priv == "argon2":
            # NOTE: Argon2 is a proper name
            der_alg = _("Argon2")
        if der_alg_priv == "argon2id":
            # NOTE: Argon2id is a proper name
            der_alg = _("Argon2id")
        elif der_alg_priv == "aeskdf":
            # NOTE: AES-KDF is a proper name
            der_alg = _("AES-KDF")

        self.derivation_algorithm_row.props.subtitle = der_alg

    def set_stats_values(self):
        self.n_entries_row.props.subtitle = str(self.entries_number)
        self.n_groups_row.props.subtitle = str(self.groups_number)
        self.n_passwords_row.props.subtitle = str(self.passwords_number)

        return GLib.SOURCE_REMOVE

    def start_stats_thread(self):
        self.entries_number = len(self.database_manager.entries)
        self.groups_number = len(self.database_manager.groups)
        self.passwords_number = 0
        for entry in self.database_manager.entries:
            if entry.password is not None and entry.password != "":
                self.passwords_number = self.passwords_number + 1
        GLib.idle_add(self.set_stats_values)

    def __on_locked(self, database_manager, _value):
        locked = database_manager.props.locked
        if locked:
            self.close()



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from collections import OrderedDict

from gi.repository import Gio, GObject


class Attribute(GObject.Object):
    __gtype_name__ = "Attribute"

    def __init__(self, key, value):
        super().__init__()

        self.key = key
        self.value = value


class AttributesModel(GObject.Object, Gio.ListModel):
    __gtype_name__ = "AttributesModel"

    def __init__(self, attributes: dict[str, str]) -> None:
        super().__init__()

        self._inner = OrderedDict()
        for key, value in attributes.items():
            self._inner[key] = Attribute(key, value)

        self.items_changed(0, 0, len(attributes))

    def do_get_item(self, pos):
        if 0 <= pos < self.get_n_items():
            key = list(self._inner)[pos]
            return self._inner[key]

        return None

    def do_get_n_items(self):
        return len(self._inner)

    def do_get_item_type(self):
        return Attribute.__gtype__  # pylint: disable=no-member

    def get(self, key: str) -> str | None:
        return self._inner.get(key)

    def _get_pos(self, key: str) -> int:
        return list(self._inner).index(key)

    def has_attribute(self, key: str) -> bool:
        return key in self._inner

    def insert(self, key: str, value: str) -> None:
        if key in self._inner:
            self._inner[key].value = value
        else:
            n_items = self.get_n_items()
            self._inner[key] = Attribute(key, value)
            self.items_changed(n_items, 0, 1)

    def pop(self, key: str) -> Attribute | None:
        if key not in self._inner:
            return None

        pos = self._get_pos(key)
        attribute = self._inner.pop(key)
        self.items_changed(pos, 1, 0)
        return attribute



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
from gettext import gettext as _
from typing import TYPE_CHECKING

import usb
import yubico
from gi.repository import Adw, Gio, GObject, Gtk

import gsecrets.config_manager as config
from gsecrets.provider.base_provider import BaseProvider

if TYPE_CHECKING:
    from gsecrets.database_manager import DatabaseManager
    from gsecrets.utils import LazyValue

NOT_INIT = "Provider not initialized."


class YubiKeyInfo(GObject.Object):
    __gtype_name__ = "YubiKeyInfo"

    def __init__(self, description: str | None = None, serial: int = 0, slot: int = 0):
        super().__init__()

        self._description = description
        self._serial = serial
        self._slot = slot
        self._raw_key = None

    @GObject.Property(type=str)
    def label(self):
        if not self._description:
            return _("No Key")

        # TRANSLATORS For example: YubiKey 4 [123456] - Slot 2
        return _("{description} [{serial}] - Slot {slot}").format(
            description=self._description,
            serial=str(self._serial),
            slot=self._slot,
        )

    @GObject.Property(type=str)
    def description(self):
        return self._description

    @GObject.Property(type=str)
    def serial(self):
        return self._serial

    @GObject.Property(type=str)
    def slot(self):
        return self._slot


class YubiKeyProvider(BaseProvider):
    def __init__(self, _window: Adw.Window) -> None:
        super().__init__()

        self.active_key: YubiKeyInfo | None = None
        self.unlock_row: Adw.ComboRow | None = None
        self.create_row: Adw.ComboRow | None = None

        # NOTE: There is currently a bug which prevents yubikey rescan
        # running under flatpak, that's why we can simply collect all yubikeys
        # here in one go and set available according to the list length
        # Needs to be fixed asap
        self.yubikeys = self.get_all_yubikeys(False)

    def get_all_yubikeys(self, debug: bool) -> list:
        """Look for YubiKeys.

        We look with ever increasing `skip' value until an error is returned.

        Return all instances of class YubiKey we got before failing.
        """
        res = []
        for _idx in range(4):
            try:
                yubikey = yubico.find_yubikey(debug=debug, skip=_idx)
            except yubico.yubikey.YubiKeyError:
                break
            else:
                try:
                    serial = yubikey.serial()
                except yubico.yubikey.YubiKeyTimeout:
                    logging.exception("Timeout getting yubikey serial")
                except yubico.yubikey.YubiKeyError:
                    logging.exception("Could not read yubikey serial")
                else:
                    logging.debug(
                        "Found %s [%s]",
                        yubikey.description,
                        str(serial),
                    )

                    res += [
                        YubiKeyInfo(yubikey.description, serial, slot)
                        for slot in yubikey.status().valid_configs()
                    ]
                finally:
                    # This must be done as otherwise usb access is broken
                    del yubikey

        return res

    def get_yubikey(self, serial: int, debug: bool = False) -> yubico.YubiKey:
        """Get a specific yubikey based on it's serial."""
        try:
            for _idx in range(4):
                yubikey = yubico.find_yubikey(debug=debug, skip=_idx)

                if yubikey.serial() == serial:
                    return yubikey

                del yubikey
        except yubico.yubikey.YubiKeyError:
            pass

        return None

    @property
    def available(self):
        return len(self.yubikeys) != 0

    def _create_model(self):
        model = Gio.ListStore.new(YubiKeyInfo)
        model.append(YubiKeyInfo())

        for this in self.yubikeys:
            model.append(this)

        return model

    def create_unlock_widget(self, database_manager: DatabaseManager) -> Gtk.Widget:
        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_factory_setup)
        factory.connect("bind", self._on_factory_bind)

        row = Adw.ComboRow()
        row.set_title(_("YubiKey"))
        row.set_subtitle(_("Select key"))
        row.set_factory(factory)
        row.connect("notify::selected", self._on_unlock_row_selected)
        self.unlock_row = row

        refresh_button = Gtk.Button()
        refresh_button.set_valign(Gtk.Align.CENTER)
        refresh_button.add_css_class("flat")
        refresh_button.set_icon_name("view-refresh-symbolic")
        # TRANSLATORS YubiKey is a proper name key, see https://en.wikipedia.org/wiki/YubiKey.
        refresh_button.set_tooltip_text(_("Select YubiKey slot"))
        refresh_button.connect("clicked", self._on_refresh_button_clicked)
        row.add_suffix(refresh_button)

        self._on_refresh_button_clicked(row)
        row.set_selected(0)

        if (
            cfg := config.get_provider_config(database_manager.path, "YubiKeyProvider")
        ) and "serial" in cfg:
            model = row.get_model()

            for pos, info in enumerate(model):
                if info.serial == cfg["serial"] and info.slot == cfg["slot"]:
                    row.set_selected(pos)
                    break

        return row

    def _on_unlock_row_selected(
        self,
        widget: Adw.ComboRow,
        _param: GObject.ParamSpec,
    ) -> None:
        self.active_key = widget.get_selected_item()

    def _on_factory_setup(self, _factory, list_item):
        label = Gtk.Label()
        label.set_xalign(0.0)
        list_item.set_child(label)

    def _on_factory_bind(self, _factory, list_item):
        label = list_item.get_child()
        info = list_item.get_item()
        label.props.label = info.props.label

    def _on_refresh_button_clicked(self, _row: Adw.ComboRow) -> None:
        model = self._create_model()

        if self.unlock_row is None:
            raise ValueError(NOT_INIT)

        self.unlock_row.set_model(model)
        self.unlock_row.set_selected(len(model) - 1)

    def create_database_row(self) -> None:
        factory = Gtk.SignalListItemFactory()
        factory.connect("setup", self._on_factory_setup)
        factory.connect("bind", self._on_factory_bind)

        self.create_row = Adw.ComboRow()
        self.create_row.set_title(_("YubiKey"))
        self.create_row.set_factory(factory)
        self.create_row.connect("notify::selected", self._on_create_row_selected)

        refresh_button = Gtk.Button()
        refresh_button.set_valign(Gtk.Align.CENTER)
        refresh_button.add_css_class("flat")
        refresh_button.set_icon_name("view-refresh-symbolic")
        refresh_button.set_tooltip_text(_("Select YubiKey slot"))
        refresh_button.connect(
            "clicked",
            self._on_yubikey_create_refresh_button_clicked,
        )
        self.create_row.add_suffix(refresh_button)

        self._on_yubikey_create_refresh_button_clicked(self.create_row)

        return self.create_row

    def _on_create_row_selected(
        self,
        widget: Adw.ComboRow,
        _param: GObject.ParamSpec,
    ) -> None:
        self.active_key = widget.get_selected_item()

    def _on_yubikey_create_refresh_button_clicked(self, _button: Gtk.Button) -> None:
        model = self._create_model()

        if self.create_row is None:
            raise ValueError(NOT_INIT)

        self.create_row.set_model(model)
        self.create_row.set_selected(0)

    def generate_key(self, salt: LazyValue[bytes]) -> bool:
        if self.active_key is None:
            return False

        try:
            if yubikey := self.get_yubikey(self.active_key.serial):
                self.emit(self.show_message, _("Touch YubiKey"))
                try:
                    self.raw_key = yubikey.challenge_response(
                        salt.value,
                        slot=self.active_key.slot,
                    )
                except yubico.yubikey_base.YubiKeyTimeout as ex:
                    self.emit(self.hide_message)
                    logging.debug("Timeout waiting for challenge response: %s", ex)
                    msg = "Timeout waiting for challenge response"
                    raise ValueError(msg) from ex

                self.emit(self.hide_message)

                # This must be done as otherwise usb access is broken
                del yubikey

                if self.raw_key is not None:
                    return True

        except usb.core.USBError as ex:
            logging.warning("USB error during yubikey key generation")
            msg = "USB error during yubikey key generation"
            raise ValueError(msg) from ex

        return False

    def config(self) -> dict:
        if self.active_key is None:
            raise ValueError(NOT_INIT)

        return {"serial": self.active_key.serial, "slot": self.active_key.slot}



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import atexit
import logging
from gettext import gettext as _
from typing import TYPE_CHECKING

from gi.repository import Adw, Gio, GLib, GObject, Gtk

# pylint: disable=import-error
from PyKCS11 import (
    CKA_CLASS,
    CKA_KEY_TYPE,
    CKA_LABEL,
    CKA_SIGN,
    CKK_RSA,
    CKO_PRIVATE_KEY,
    CKR_USER_ALREADY_LOGGED_IN,
    CKS_RW_USER_FUNCTIONS,
    PyKCS11,
    PyKCS11Lib,
)

import gsecrets.config_manager as config
from gsecrets import const
from gsecrets.provider.base_provider import BaseProvider

if TYPE_CHECKING:
    from gsecrets.database_manager import DatabaseManager
    from gsecrets.utils import LazyValue

# KeePass-Smart-Certificate-Key-Provider
# https://github.com/BodnarSoft/KeePass-Smart-Certificate-Key-Provider/
# Source/SmartCertificateKeyProvider.cs
# Static string: do not change as otherwise we are no longer compatible
KEEPASS_SCKP_TEXT = (
    "Data text for KeePass Password Safe Plugin"
    " - {F3EF424C-7517-4D58-A3FB-C1FB458FDDB6}!"
)


class Pkcs11Provider(BaseProvider):
    def __init__(self, window):
        super().__init__()
        self.window = window

        self._pkcs11 = None
        self._session = None

        self._active_certificate = None

        atexit.register(self._cleanup)

    def _cleanup(self):
        if self._session:
            try:
                self._session.logout()
            except PyKCS11.PyKCS11Error:
                logging.exception("Could not cleanup")

            self._session.closeSession()

    def scan_slots(self):
        slot_list = self._pkcs11.getSlotList(tokenPresent=True)
        if slot_list:
            self._session = self._pkcs11.openSession(
                slot_list[0],
                PyKCS11.CKF_RW_SESSION,
            )
            return True

        return False

    @property
    def available(self):
        return True

    def logout(self):
        try:
            self._session.logout()
        except PyKCS11.PyKCS11Error:
            logging.exception("Could not logout")

    def login(self, pin: str) -> bool:
        try:
            self._session.login(pin)
        except PyKCS11.PyKCS11Error as err:
            # Check if we were already logged in (due to some race condition)
            if err.value != CKR_USER_ALREADY_LOGGED_IN:
                logging.exception("Could not login")
                return False

        logging.debug("Successfully logged in")
        return True

    def _create_model(self):
        model = Gtk.StringList()
        model.append(_("No Smartcard"))

        if not self._pkcs11:
            return model

        if not self.scan_slots():
            return model

        model.splice(0, 1, [_("No RSA Certificate")])
        objects = self._session.findObjects(
            [(CKA_CLASS, CKO_PRIVATE_KEY), (CKA_KEY_TYPE, CKK_RSA), (CKA_SIGN, True)],
        )
        for obj in objects:
            value = self._session.getAttributeValue(obj, [CKA_LABEL])[0]
            model.append(value)

        return model

    def create_unlock_widget(self, database_manager: DatabaseManager) -> Gtk.Widget:
        row = Adw.ComboRow()
        row.set_title(_("Smartcard"))
        row.set_subtitle(_("Select certificate"))
        row.connect("notify::selected", self._on_unlock_row_selected)

        self.refresh_stack = Gtk.Stack()
        self.refresh_button = get_refresh_button()
        self.refresh_button.connect("clicked", self._on_refresh_button_clicked, row)
        self.refresh_stack.add_named(self.refresh_button, "button")

        self.refresh_spinner = Adw.Spinner()
        self.refresh_stack.add_named(self.refresh_spinner, "spinner")
        row.add_suffix(self.refresh_stack)
        self.refresh_stack.set_visible_child(self.refresh_button)

        self.fill_data(row, database_manager)

        row.set_selected(0)

        return row

    def _on_unlock_row_selected(
        self,
        widget: Adw.ComboRow,
        _param: GObject.ParamSpec,
    ) -> None:
        if selected_item := widget.get_selected_item():
            self._active_certificate = selected_item.get_string()

    def _refresh_pkcs11_async(
        self,
        callback: Gio.AsyncReadyCallback,
        row: Adw.ComboRow,
    ) -> None:
        def pkcs11_refresh(task, _source_object, _task_data, _cancellable):
            self._pkcs11 = None
            if not self._pkcs11:
                self._pkcs11 = PyKCS11Lib()

                try:
                    self._pkcs11.load(const.PKCS11_LIB)
                except PyKCS11.PyKCS11Error as err:
                    logging.warning("Could not load pkcs11 library: %s", err)
                    task.return_error(err)
                    return

            present = self.scan_slots()
            task.return_boolean(present)

        self.refresh_stack.set_visible_child(self.refresh_spinner)

        task = Gio.Task.new(self, None, callback, row)
        task.run_in_thread(pkcs11_refresh)

    def _refresh_pkcs11_finish(self, result):
        self.refresh_stack.set_visible_child(self.refresh_button)
        return result.propagate_boolean()

    def _on_refresh_button_clicked(
        self,
        _button: Gtk.Button,
        row: Adw.ComboRow,
    ) -> None:
        self._refresh_pkcs11_async(self._on_pkcs11_refresh, row)

    def _on_pkcs11_refresh(
        self,
        _provider: Pkcs11Provider,
        result: Gio.AsyncResult,
        row: Adw.ComboRow,
    ) -> None:
        try:
            present = self._refresh_pkcs11_finish(result)
        except GLib.Error:
            present = False

        if not present:
            dialog = Adw.AlertDialog.new(
                _("No smartcard present"),
                _("Please insert smartcard and retry."),
            )

            dialog.add_response("ok", _("_OK"))
            dialog.present(self.window)
            return

        session_info = self._session.getSessionInfo()
        if session_info.state >= CKS_RW_USER_FUNCTIONS:
            # We are already authorized to perform R/W user functions.
            # This means that we are already considered 'logged in' and
            # do not need to ask for a password.
            logging.debug("Already logged in")
            self.fill_data(row)
            return

        entry = Adw.PasswordEntryRow(activates_default=True, title=_("Passphrase"))
        entry.add_css_class("card")

        dialog = Adw.AlertDialog.new(_("Unlock"), _("Unlock your smartcard"))
        dialog.add_response("cancel", _("_Cancel"))
        dialog.add_response("unlock", _("_Unlock"))
        dialog.set_default_response("unlock")
        dialog.set_extra_child(entry)
        dialog.set_focus(entry)
        dialog.connect("response", self._on_pin_dialog_response, entry, row)
        dialog.present(self.window)

    def fill_data(
        self,
        row: Adw.ComboRow,
        database_manager: DatabaseManager | None = None,
    ) -> None:
        model = self._create_model()
        row.set_model(model)

        if not database_manager:
            return

        row_select = 0
        if (
            cfg := config.get_provider_config(database_manager.path, "Pkcs11Provider")
        ) and "serial" in cfg:
            model = row.get_model()

            for pos, info in enumerate(model):
                if info == cfg["serial"]:
                    row_select = pos
                    break

        row.set_selected(row_select)

    def _on_pin_dialog_response(
        self,
        _dialog: Adw.AlertDialog,
        response: str,
        entry: Gtk.Entry,
        row: Adw.ComboRow,
    ) -> None:
        if response == "unlock":
            pin = entry.get_text()

            ret = self.login(pin)
            if not ret:
                self.window.send_notification(_("Failed to unlock Smartcard"))
                return

            self.fill_data(row)

    def create_database_row(self):
        row = Adw.ComboRow()
        row.set_title(_("Smartcard"))
        row.set_subtitle(_("Use a smartcard"))
        row.connect("notify::selected", self._on_unlock_row_selected)

        self.refresh_stack = Gtk.Stack()
        self.refresh_button = get_refresh_button()
        self.refresh_button.connect("clicked", self._on_refresh_button_clicked, row)
        self.refresh_stack.add_named(self.refresh_button, "button")

        self.refresh_spinner = Adw.Spinner()
        self.refresh_stack.add_named(self.refresh_spinner, "spinner")
        row.add_suffix(self.refresh_stack)
        self.refresh_stack.set_visible_child(self.refresh_button)

        self.fill_data(row)

        row.set_selected(0)

        return row

    def generate_key(self, _salt: LazyValue[bytes]) -> bool:
        if not self._session:
            return False

        objs = self._session.findObjects(
            [(CKA_CLASS, CKO_PRIVATE_KEY), (CKA_LABEL, self._active_certificate)],
        )
        if len(objs) == 0:
            return False

        priv_key = objs[0]
        mecha = PyKCS11.Mechanism(PyKCS11.CKM_SHA1_RSA_PKCS, None)

        try:
            signed = self._session.sign(priv_key, KEEPASS_SCKP_TEXT, mecha)
        except PyKCS11.PyKCS11Error as err:
            logging.exception("Could not sign data, abort")
            msg = "Could not sign data"
            raise ValueError(msg) from err

        signed_bytes = bytearray(signed)
        immutable_bytes = bytes(signed_bytes)

        self.raw_key = immutable_bytes

        return True

    def config(self) -> dict:
        return {"label": self._active_certificate}


def get_refresh_button():
    button = Gtk.Button.new_from_icon_name("view-refresh-symbolic")
    button.set_valign(Gtk.Align.CENTER)
    button.add_css_class("flat")
    button.set_tooltip_text(_("Refresh Certificate List"))
    return button



# SPDX-License-Identifier: GPL-3.0-only
"""Generate a passphrase from a list of predefined words."""

from __future__ import annotations

import logging
from secrets import randbelow

from gi.repository import Gio, GLib, GObject


class Passphrase(GObject.Object):
    """Generate a passphrase from a list of predefined words."""

    _word_file_data = None

    def generate(self, num_words: int, separator: str = "-") -> None:
        """Generate a passphrase.

        :param int num_words: number of words requested
        :param str separator: separator
        """
        if data := self._word_file_data:
            self._parse_data(data)
            return

        word_file: Gio.File = Gio.File.new_for_uri(
            "resource:///org/gnome/World/Secrets/crypto/eff_large_wordlist.txt",
        )

        def callback(gfile: Gio.File, result: Gio.AsyncResult) -> None:
            try:
                gbytes, _ = gfile.load_bytes_finish(result)
            except GLib.Error:
                logging.exception("Could not read word file")
            else:
                self._word_file_data = gbytes
                self._parse_data(gbytes, num_words, separator)

        word_file.load_bytes_async(None, callback)

    @GObject.Signal(arg_types=(str,))
    def generated(self, _passphrase):
        return

    def _parse_data(
        self,
        gbytes: GLib.Bytes,
        num_words: int,
        separator: str = "-",
    ) -> None:
        word_str: str = gbytes.get_data().decode("utf-8")
        word_list: list[str] = word_str.split("\n")
        len_words: int = len(word_list)

        words = [word_list[randbelow(len_words)] for _ in range(num_words)]
        passphrase: str = separator.join(words) or ""
        self.emit("generated", passphrase)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gi.repository import Adw, GLib, GObject, Gtk

DURATION = 3.0


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/error_revealer.ui")
class ErrorRevealer(Adw.Bin):
    __gtype_name__ = "ErrorRevealer"

    _revealer = Gtk.Template.Child()

    label = GObject.Property(type=str, default="")
    _source_id: int | None = None

    def reveal(self, reveal: bool, label: str | None = None) -> None:
        if self._source_id:
            GLib.source_remove(self._source_id)
            self._source_id = None

        if reveal:
            if label:
                self.props.label = label

            self._revealer.props.reveal_child = True
            self._source_id = GLib.timeout_add_seconds(DURATION, self._hide_callback)
        else:
            self._revealer.props.reveal_child = False

    def _hide_callback(self) -> None:
        self.reveal(False)

        return GLib.SOURCE_REMOVE



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gi.repository import Adw, Gio, Gtk

from gsecrets.safe_element import SafeElement
from gsecrets.widgets.history_row import HistoryRow


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/history_window.ui")
class HistoryWindow(Adw.Dialog):
    __gtype_name__ = "HistoryWindow"

    listbox = Gtk.Template.Child()

    def __init__(self, safe_entry, unlocked_database):
        super().__init__()

        self.list_model = Gio.ListStore.new(SafeElement)
        self.safe_entry = safe_entry
        self.unlocked_database = unlocked_database

        history = safe_entry.history[::-1]  # Reversed, order by newest first.
        self.list_model.splice(0, 0, history)
        self.listbox.bind_model(self.list_model, HistoryRow, self)

        self._signal_handle = self.unlocked_database.database_manager.connect(
            "notify::locked",
            self._on_locked,
        )

    def _on_locked(self, database_manager, _value):
        if database_manager.props.locked:
            self.close()

    def do_closed(self) -> None:
        if handle := self._signal_handle:
            self.unlocked_database.database_manager.disconnect(handle)
            self._signal_handle = 0



# SPDX-License-Identifier: GPL-3.0-only
from gi.repository import GObject


class BaseProvider(GObject.Object):
    __gtype_name__ = "BaseProvider"

    show_message = GObject.Signal(arg_types=(str,))
    hide_message = GObject.Signal(arg_types=())

    def __init__(self):
        super().__init__()

        self.raw_key = None

    @property
    def available(self) -> bool:
        return False

    @property
    def key(self) -> bytes:
        return self.raw_key

    def config(self) -> dict:
        return {}

    def clear_input_fields(self) -> None:
        pass



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gi.repository import Adw, GObject, Gtk


class AddListBox(Adw.Bin):
    """ListBox with a Add button at the end."""

    __gtype_name__ = "AddListBox"

    _n_items: int = 0

    row_activated = GObject.Signal(arg_types=(Gtk.ListBoxRow,))

    def __init__(self):
        super().__init__()

        self._inner = Gtk.ListBox(selection_mode=Gtk.SelectionMode.NONE)
        self._inner.add_css_class("boxed-list")
        self._inner.connect("row-activated", self._on_row_activated)

        add_more_row = Adw.ButtonRow()
        add_more_row.props.start_icon_name = "list-add-symbolic"
        add_more_row.props.use_underline = True

        self.set_child(self._inner)
        self._inner.append(add_more_row)

        self._row = add_more_row
        self._model = None

    def _on_row_activated(self, _list_box, row):
        self.emit(self.row_activated, row)

    @GObject.Property(type=str, default="")
    def action_name(self):
        return self._row.props.action_name

    @action_name.setter  # type: ignore
    def action_name(self, action_name):
        self._row.props.action_name = action_name

    @GObject.Property(type=str, default="")
    def label(self):
        return self._row.props.title

    @label.setter  # type: ignore
    def label(self, string):
        self._row.props.title = string

    def append(self, child):  # pylint: disable=arguments-differ
        self._inner.insert(child, self._n_items)
        self._n_items += 1

    def remove(self, child):  # pylint: disable=arguments-differ
        if child.props.parent != self._inner:
            return

        self._inner.remove(child)
        self._n_items -= 1

    def _insert(self, pos, child):  # pylint: disable=arguments-differ
        self._inner.insert(child, pos)
        self._n_items += 1

    def set_model(self, model, factory):
        self._model = model

        def on_items_changed(model, pos, removed, added):
            for idx in range(pos, pos + removed):
                if child := self._inner.get_row_at_index(idx):
                    self.remove(child)

            for idx in range(pos, pos + added):
                item = model.get_item(idx)
                self._insert(idx, factory(item))

        on_items_changed(model, 0, 0, model.get_n_items())
        model.connect("items-changed", on_items_changed)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gi.repository import Adw, GObject, Gtk

from gsecrets.safe_element import SafeGroup
from gsecrets.widgets.notes_dialog import NotesDialog


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/group_page.ui")
class GroupPage(Adw.Bin):
    __gtype_name__ = "GroupPage"

    title_entry_row = Gtk.Template.Child()
    notes_text_view = Gtk.Template.Child()
    _window_title = Gtk.Template.Child()

    safe_group = GObject.Property(type=SafeGroup)

    def __init__(self, unlocked_database, safe_group):
        super().__init__(safe_group=safe_group)

        self.unlocked_database = unlocked_database

        safe_group = self.props.safe_group

        # Setup Widgets
        notes_buffer = self.notes_text_view.get_buffer()

        # Connect Signals
        safe_group.updated.connect(self._on_safe_group_updated)
        safe_group.bind_property(
            "name",
            self.title_entry_row,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )
        safe_group.bind_property(
            "notes",
            notes_buffer,
            "text",
            GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.BIDIRECTIONAL,
        )

        safe_group.connect("notify::name", self._on_name_notify)
        self._set_title(safe_group)

    def _set_title(self, group):
        if group.parentgroup.is_root_group:
            title = group.name
        else:
            title = f"{group.parentgroup.name}  ⟩ {group.name}"

        self._window_title.props.title = title

    def _on_name_notify(self, group, _pspec):
        self._set_title(group)

    @Gtk.Template.Callback()
    def on_notes_detach_button_clicked(self, _button):
        self.unlocked_database.start_database_lock_timer()
        safe_group = self.props.safe_group
        NotesDialog(self.unlocked_database, safe_group).present(self)

    def _on_safe_group_updated(self, _safe_group: SafeGroup) -> None:
        self.unlocked_database.start_database_lock_timer()



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import logging
from enum import IntEnum
from gettext import gettext as _

from gi.repository import Adw, Gio, GLib, Gtk

import gsecrets.config_manager
from gsecrets import const
from gsecrets.create_database import CreateDatabase
from gsecrets.database_manager import DatabaseManager
from gsecrets.err import QUARK, ErrorType
from gsecrets.provider.providers import Providers
from gsecrets.recent_manager import RecentManager
from gsecrets.save_dialog import SaveDialog
from gsecrets.settings_dialog import SettingsDialog
from gsecrets.unlock_database import UnlockDatabase
from gsecrets.widgets.quit_conflict_dialog import QuitConflictDialog


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/window.ui")
class Window(Adw.ApplicationWindow):
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-many-public-methods
    class View(IntEnum):
        WELCOME = 0
        UNLOCK_DATABASE = 2
        UNLOCKED_DATABASE = 3
        CREATE_DATABASE = 4

    __gtype_name__ = "Window"

    _create_view = None
    unlocked_db = None

    _create_database_bin = Gtk.Template.Child()
    _main_view = Gtk.Template.Child()
    toast_overlay = Gtk.Template.Child()
    _unlock_database_bin = Gtk.Template.Child()
    unlocked_db_bin = Gtk.Template.Child()

    _view = View.WELCOME

    def __init__(self, *args, **kwargs):
        self.install_action("win.new_database", None, self.on_new_database_action)

        super().__init__(*args, **kwargs)

        self.key_providers = Providers(self)

        for provider in self.key_providers.get_key_providers():
            if not provider.available:
                continue

            provider.connect(provider.show_message, self._on_show_message)
            provider.connect(provider.hide_message, self._on_hide_message)

        self.application = self.get_application()

        self.assemble_window()
        self.setup_actions()

        if self.application.development_mode is True:
            gsecrets.config_manager.set_development_backup_mode(True)

    def send_notification(self, notification: str) -> None:
        toast = Adw.Toast.new(notification)
        self.toast_overlay.add_toast(toast)

    def _on_show_message(self, _provider, label):
        self.show_banner(label)

    def _on_hide_message(self, _provider):
        self.close_banner()

    def show_banner(self, label: str) -> None:
        if self.view == self.View.UNLOCK_DATABASE:
            self._unlock_database_bin.props.child.show_banner(label)
        elif self.view == self.View.CREATE_DATABASE and (
            create_view := self._create_view
        ):
            create_view.show_banner(label)

    def close_banner(self):
        if self.view == self.View.UNLOCK_DATABASE:
            self._unlock_database_bin.props.child.close_banner()
        elif self.view == self.View.CREATE_DATABASE and (
            create_view := self._create_view
        ):
            create_view.close_banner()

    def assemble_window(self) -> None:
        window_size = gsecrets.config_manager.get_window_size()
        self.set_default_size(window_size[0], window_size[1])

        if self.props.application.development_mode:
            self.add_css_class("devel")

    def do_enable_debugging(  # pylint: disable=arguments-differ
        self,
        toggle: bool,
    ) -> bool:
        if not self.application.development_mode:
            logging.warning("The inspector is not enabled in non-development builds")
            return False

        return Adw.ApplicationWindow.do_enable_debugging(self, toggle)

    def invoke_initial_screen(self) -> None:
        """Present the first start screen if required or autoload files.

        If the configuration is set to automatically load the last
        opened safe, this function does that. If it is not set to
        autoload, it presents a list of recently loaded files (or
        displays the empty welcome page if there are no recently
        loaded files).
        """
        if gsecrets.config_manager.get_first_start_screen():
            # simply load the last opened file
            uri = gsecrets.config_manager.get_last_opened_database()
            gfile: Gio.File = Gio.File.new_for_uri(uri)
            recents = RecentManager()
            recent_uris = [f.get_uri() for f in recents]
            if gfile.query_exists() and uri in recent_uris:
                filepath = gfile.get_path()
                logging.debug("Opening last opened database: %s", filepath)
                self.start_database_opening_routine(filepath)
                return

        self.view = self.View.WELCOME

    #
    # Open Database Methods
    #

    def on_open_database_action(self, _action: Gio.Action, param: GLib.Variant) -> None:
        """Callback function to open a safe."""
        # pylint: disable=too-many-locals
        if param and (path := param.get_string()):
            self._open_database_in_window(path)
            return

        keepass_filter = Gtk.FileFilter()
        keepass_filter.add_mime_type("application/x-keepass2")
        keepass_filter.add_pattern("*.kdbx")
        # TRANSLATORS Safe as in strongbox.
        keepass_filter.set_name("KeePass 3.1/4 " + _("Safe"))

        binary_filter = Gtk.FileFilter()
        binary_filter.add_mime_type("application/octet-stream")
        binary_filter.set_name(_("Any file type"))

        filters = Gio.ListStore.new(Gtk.FileFilter)
        filters.append(keepass_filter)
        filters.append(binary_filter)

        dialog = Gtk.FileDialog.new()
        dialog.props.title = _("Select Safe")
        dialog.props.filters = filters

        dialog.open(self, None, self._on_open_filechooser_response)

    def _on_open_filechooser_response(self, dialog, result):
        try:
            db_gfile = dialog.open_finish(result)
        except GLib.Error as err:
            if not err.matches(Gtk.DialogError.quark(), Gtk.DialogError.DISMISSED):
                logging.exception("Could not open file")
        else:
            if db_filename := db_gfile.get_path():
                logging.debug("File selected: %s", db_filename)
                self._open_database_in_window(db_filename)
            else:
                logging.error("File does not have a path, is in a virtual file system?")
                self.send_notification(_("File does not have a valid path"))

    def _open_database_in_window(self, filepath):
        if self.unlocked_db:
            auto_save = gsecrets.config_manager.get_save_automatically()
            is_dirty = self.unlocked_db.database_manager.is_dirty
            is_current = self.unlocked_db.database_manager.path == filepath
            is_locked = self.unlocked_db.database_manager.props.locked

            if is_locked:
                if is_dirty and not is_current:
                    app = self.props.application
                    window = app.new_window()
                    window.start_database_opening_routine(filepath)
                    window.present()
                else:
                    if is_current:
                        self.view = self.View.UNLOCK_DATABASE
                        return

                    if auto_save:
                        self.unlocked_db.auto_save_database()

                    self.unlocked_db.inner_dispose()
                    self.unlocked_db = None
                    self.start_database_opening_routine(filepath)

            elif not is_current:
                app = self.props.application
                window = app.new_window()
                window.start_database_opening_routine(filepath)
                window.present()
            else:
                self.send_notification(_("Safe is already open"))

        else:
            self.start_database_opening_routine(filepath)

    def start_database_opening_routine(self, filepath: str) -> None:
        """Start opening a safe file."""
        # TODO Use a Gio.File instead of a path.
        database = Gio.File.new_for_path(filepath)
        unlock_db = UnlockDatabase(self, database)
        self._unlock_database_bin.props.child = unlock_db
        self.view = self.View.UNLOCK_DATABASE
        unlock_db.grab_entry_focus()

    #
    # Create Database Methods
    #

    def on_new_database_action(
        self, _widget: Gtk.Widget, _action_name: str, _parameter: GLib.Variant
    ) -> None:
        """Callback function to create a new safe."""
        filter_text = Gtk.FileFilter()
        # TRANSLATORS Safe as in strongbox.
        filter_text.set_name("KeePass 3.1/4 " + _("Safe"))
        filter_text.add_mime_type("application/x-keepass2")

        filters = Gio.ListStore.new(Gtk.FileFilter)
        filters.append(filter_text)

        dialog = Gtk.FileDialog.new()
        # NOTE: Filechooser title for creating a new keepass safe kdbx file
        dialog.props.title = _("Create Safe")
        dialog.props.filters = filters
        dialog.props.accept_label = _("_Create")
        # NOTE: Safe as in strongbox
        dialog.props.initial_name = _("Safe") + ".kdbx"

        dialog.save(self, None, self._on_create_filechooser_response)

    def _on_create_filechooser_response(self, dialog, result):
        try:
            gfile = dialog.save_finish(result)
        except GLib.Error as err:
            if not err.matches(Gtk.DialogError.quark(), Gtk.DialogError.DISMISSED):
                logging.exception("Could not save file")
        else:
            filepath = gfile.get_path()
            if self.unlocked_db:
                auto_save = gsecrets.config_manager.get_save_automatically()
                is_dirty = self.unlocked_db.database_manager.is_dirty
                is_open = self.application.is_safe_open(filepath)
                is_locked = self.unlocked_db.database_manager.props.locked

                if is_open:
                    self.send_notification(
                        _("Cannot create Safe: Safe is already open"),
                    )
                    return

                if is_locked:
                    if is_dirty and not auto_save:
                        app = self.props.application
                        window = app.new_window()
                        window.create_new_database(filepath)
                        window.present()
                    else:
                        if auto_save:
                            self.unlocked_db.auto_save_database()

                        self.unlocked_db.inner_dispose()
                        self.unlocked_db = None
                        self.create_new_database(filepath)
                else:
                    app = self.props.application
                    window = app.new_window()
                    window.create_new_database(filepath)
                    window.present()

            else:
                self.create_new_database(filepath)

    def create_new_database(self, filepath: str) -> None:
        """Copy stock database onto filepath."""
        self._main_view.set_visible_child_name("spinner")

        stock_db_file: Gio.File = Gio.File.new_for_uri(
            "resource:///org/gnome/World/Secrets/database.kdbx",
        )
        new_db_file: Gio.File = Gio.File.new_for_path(filepath)

        def unlock_callback(database_manager, result):
            try:
                database_manager.unlock_finish(result)
            except GLib.Error as err:
                if not err.matches(QUARK, ErrorType.CREDENTIALS_ERROR):
                    logging.exception("Could not unlock safe")

                self.invoke_initial_screen()
                self.send_notification(_("Could not create new Safe"))
            else:
                create_database = CreateDatabase(self, database_manager)
                self._create_database_bin.props.child = create_database
                self._create_view = create_database
                self.view = self.View.CREATE_DATABASE

        def copy_callback(gfile, result):
            try:
                gfile.copy_finish(result)
            except GLib.Error:
                logging.exception("Could not copy new database")
                self.invoke_initial_screen()
                self.send_notification(_("Could not create new Safe"))
            else:
                database_manager = DatabaseManager(self.key_providers, filepath)
                database_manager.unlock_async(
                    "liufhre86ewoiwejmrcu8owe",
                    callback=unlock_callback,
                )

        if GLib.check_version(2, 81, 0) is None:
            stock_db_file.copy_async(
                new_db_file,
                Gio.FileCopyFlags.OVERWRITE,
                GLib.PRIORITY_DEFAULT,
                None,
                None,
                copy_callback,
            )
        else:
            stock_db_file.copy_async(
                new_db_file,
                Gio.FileCopyFlags.OVERWRITE,
                GLib.PRIORITY_DEFAULT,
                None,
                None,
                None,
                copy_callback,
            )

    def save_window_size(self):
        width = self.get_width()
        height = self.get_height()
        logging.debug("Saving window geometry: (%s, %s)", width, height)
        gsecrets.config_manager.set_window_size([width, height])

    def force_close(self):
        """Method to close a window with unsaved changes.

        Should only be called from SaveDialog and QuitConflictDialog.
        """
        if unlocked_db := self.unlocked_db:
            unlocked_db.database_manager.is_dirty = False

        self.close()

    def do_close_request(self) -> bool:  # pylint: disable=arguments-differ
        """Invoked when the window close button is pressed.

        Only the app.quit action is called. This action cleans up stuff
        and will invoke the on_application_shutdown() method.

        Return: True to stop other handlers from being invoked for the signal.
        """
        if not self.unlocked_db:
            self.save_window_size()
            return False

        dbm = self.unlocked_db.database_manager
        is_database_dirty = dbm.is_dirty

        def on_save(dbm, result):
            try:
                is_saved = dbm.save_finish(result)
            except GLib.Error:
                logging.exception("Could not save Safe")
                self.show_quit_confirmation_dialog()
            else:
                if is_saved:
                    self.close()
                else:
                    # This shouldn't happen
                    self.show_quit_confirmation_dialog()

        def on_check_file_changes(dbm, result):
            try:
                conflicts = dbm.check_file_changes_finish(result)
            except GLib.Error:
                logging.exception("Could not monitor file changes")
                self.send_notification(_("Could not save safe"))
            else:
                if conflicts:
                    dialog = QuitConflictDialog(self, dbm, on_save)
                    dialog.present()
                elif gsecrets.config_manager.get_save_automatically():
                    self.unlocked_db.database_manager.save_async(on_save)
                else:
                    save_dialog = SaveDialog(self)
                    save_dialog.present(self)

        if is_database_dirty:
            dbm.check_file_changes_async(on_check_file_changes)
            return True

        self.save_window_size()
        return False

    def show_quit_confirmation_dialog(self):
        def on_discard(self, _dialog, _response):
            self.force_close()

        # TODO Set the body of the message based on the error kind.
        dialog = Adw.AlertDialog.new(_("Could not save Safe"), None)
        dialog.add_response("discard", _("_Quit Without Saving"))
        dialog.add_response("cancel", _("_Don't Quit"))
        dialog.set_response_appearance("discard", Adw.ResponseAppearance.DESTRUCTIVE)
        dialog.set_default_response("cancel")
        dialog.connect("response::discard", on_discard)
        dialog.present(self)

    def do_unrealize(self):  # pylint: disable=arguments-differ
        if self.unlocked_db:
            self.unlocked_db.cleanup()

        Gtk.Widget.do_unrealize(self)

    def setup_actions(self):
        sort_action = self.application.settings.create_action("sort-order")
        self.add_action(sort_action)

        selection_action = Gio.SimpleAction.new("db.selection", GLib.VariantType("s"))
        self.add_action(selection_action)
        selection_action.connect("activate", self.execute_database_action)

        about_action = Gio.SimpleAction.new("about", None)
        self.add_action(about_action)
        about_action.connect("activate", self.on_about_action)

        settings_action = Gio.SimpleAction.new("settings", None)
        self.add_action(settings_action)
        settings_action.connect("activate", self.on_settings_action)

        def on_visible_dialog_notify(window, _pspec):
            enabled = window.props.visible_dialog is None
            go_back_action.set_enabled(enabled)

        go_back_action = Gio.SimpleAction.new("go_back", None)
        self.add_action(go_back_action)
        go_back_action.connect("activate", self.on_go_back_action)
        # NOTE We need to do this to avoid consuming Esc events
        self.connect("notify::visible-dialog", on_visible_dialog_notify)

        open_database_action = Gio.SimpleAction.new(
            "open_database",
            GLib.VariantType("s"),
        )
        self.add_action(open_database_action)
        open_database_action.connect("activate", self.on_open_database_action)

        actions = [
            "db.save",
            "db.save_dirty",
            "db.lock",
            "db.add_entry",
            "db.add_group",
            "db.settings",
            "db.undo_delete",
            "db.undo_attribute_delete",
            "element.delete",
            "entry.duplicate",
            "entry.references",
            "element.properties",
        ]

        # The save_dirty action differs to save in that it is set as disabled
        # by default, and it is used by the main menu to determine if the button
        # should be set to sensitive.
        for action in actions:
            simple_action = Gio.SimpleAction.new(action, None)
            if action == "db.save_dirty":
                simple_action.set_enabled(False)

            simple_action.connect("activate", self.execute_database_action)
            self.add_action(simple_action)

    def execute_database_action(self, action, param):
        # pylint: disable=too-many-branches
        action_db = self.unlocked_db
        if action_db is None:
            return

        action_db.start_database_lock_timer()
        name = action.props.name

        # We don't want to allow actions if the db is locked.
        # TODO Actions should be disabled instead.
        if name == "db.lock":
            action_db.lock_safe()
        elif action_db.props.database_locked:
            return

        if name == "element.delete":
            action_db.on_element_delete_action()
        elif name == "entry.duplicate":
            action_db.on_entry_duplicate_action()
        elif name == "entry.references":
            action_db.show_references_dialog()
        elif name == "element.properties":
            action_db.show_properties_dialog()
        elif name == "db.settings":
            action_db.show_database_settings()
        elif name == "db.selection":
            action_db.selection_mode_headerbar.on_selection_action(param)
        elif name == "db.undo_delete":
            action_db.undo_delete()
        elif name == "db.undo_attribute_delete":
            action_db.undo_attribute_delete()
        elif name in ["db.save", "db.save_dirty"]:
            action_db.save_database()
        elif name in ["db.add_entry", "db.add_group"]:
            if action_db.props.database_locked or action_db.props.selection_mode:
                return

            if action_db.props.search_active:
                action_db.props.search_active = False

            if name == "db.add_entry":
                action_db.on_add_entry_action()
            else:
                action_db.on_add_group_action()

    def on_about_action(self, _action: Gio.Action, _param: GLib.Variant) -> None:
        """Invoked when we click "about" in the main menu."""
        about_dialog = Adw.AboutDialog.new_from_appdata(
            f"/org/gnome/World/Secrets/{const.APP_ID}.metainfo.xml",
            const.VERSION,
        )
        about_dialog.set_designers(["Christopher Davis", "Tobias Bernard"])
        about_dialog.set_developers(
            [
                "Falk Alexander Seidl <fseidl@gnome.org>",
                "Uta Lemke",
                "David Heidelberg",
                "Jean Felder",
                "Maximiliano Sandoval <msandova@gnome.org>",
                "Sebastian Spaeth",
                "Christopher Davis",
                "Jan-Michael Brummer",
            ],
        )
        about_dialog.set_copyright("© 2018-2024 Falk Alexander Seidl, et al.")
        # TRANSLATORS Add your name to the translator credits list
        about_dialog.set_translator_credits(_("translator-credits"))
        about_dialog.present(self)

    def on_settings_action(self, _action: Gio.Action, _param: GLib.Variant) -> None:
        if self.get_visible_dialog() is None:
            SettingsDialog().present(self)

    def on_go_back_action(self, _action: Gio.Action, _param: GLib.Variant) -> None:
        if self.view == self.View.CREATE_DATABASE:
            if create_view := self._create_view:
                create_view.go_back()
        elif (action_db := self.unlocked_db) and not action_db.props.database_locked:
            action_db.go_back()

    @Gtk.Template.Callback()
    def on_back_button_pressed(
        self,
        gesture: Gtk.GestureClick,
        _n_press: int,
        _event_x: float,
        _event_y: float,
    ) -> None:
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
        self.lookup_action("go_back").activate()

    @property
    def view(self) -> View:
        return self._view

    @view.setter  # type: ignore
    def view(self, new_view: View) -> None:
        stack = self._main_view

        self._view = new_view

        if new_view == self.View.WELCOME:
            stack.props.visible_child_name = "welcome"
        elif new_view == self.View.UNLOCK_DATABASE:
            stack.props.visible_child_name = "unlock_database"
        elif new_view == self.View.UNLOCKED_DATABASE:
            stack.props.visible_child_name = "unlocked"
        elif new_view == self.View.CREATE_DATABASE:
            stack.props.visible_child_name = "create_database"



# SPDX-License-Identifier: GPL-3.0-only

APP_ID = "org.gnome.World.Secrets"
VERSION = "13.0.1"
GETTEXT_PACKAGE = "secrets"
LOCALEDIR = "/usr/share/locale"
NAME = "Secrets"
PKGDATADIR = "/usr/share/secrets"
PROFILE = ""
PKCS11_LIB = "/usr/lib/opensc-pkcs11.so"

IS_DEVEL = PROFILE == "Devel"
SUFFIX = " (Devel)" if IS_DEVEL else ""
SHORT_NAME = GETTEXT_PACKAGE



# SPDX-License-Identifier: GPL-3.0-only
from gi.repository import Adw, Gtk

from gsecrets.recent_files_menu import RecentFilesMenu
from gsecrets.recent_manager import RecentManager


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/locked_headerbar.ui")
class LockedHeaderBar(Adw.Bin):
    __gtype_name__ = "LockedHeaderBar"

    title = Gtk.Template.Child()
    split_button = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

        self.recents = RecentManager()
        self.set_menu()
        self.recents.changed.connect(self.on_settings_changed)

    def set_menu(self):
        menu = RecentFilesMenu()
        if menu.is_empty:
            self.split_button.set_menu_model(None)
        else:
            self.split_button.set_menu_model(menu.menu)

    def on_settings_changed(self, _recents):
        self.set_menu()



# SPDX-License-Identifier: GPL-3.0-only

from gi.repository import Adw, Gio, Gtk

import gsecrets.config_manager as config
from gsecrets.const import APP_ID
from gsecrets.recent_manager import RecentManager


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/settings_dialog.ui")
class SettingsDialog(Adw.PreferencesDialog):
    __gtype_name__ = "SettingsDialog"

    _clear_button = Gtk.Template.Child()
    _clearcb_spin_row = Gtk.Template.Child()
    _generator_length_spin_row = Gtk.Template.Child()

    _generator_separator_entry = Gtk.Template.Child()
    _generator_words_spin_row = Gtk.Template.Child()
    _lockdb_spin_row = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

        self._set_config_values()

        delegate = self._generator_separator_entry.get_delegate()
        delegate.props.max_length = 1

    def _set_config_values(self):
        # pylint: disable=too-many-locals
        # pylint: disable=too-many-statements
        action_group = Gio.SimpleActionGroup.new()
        settings = Gio.Settings.new(APP_ID)

        # General
        first_start_action = settings.create_action("first-start-screen")
        action_group.add_action(first_start_action)

        # Safe
        save_automatically_action = settings.create_action("save-automatically")
        action_group.add_action(save_automatically_action)

        # Password Generator
        settings.bind(
            "generator-length",
            self._generator_length_spin_row,
            "value",
            Gio.SettingsBindFlags.DEFAULT,
        )

        use_uppercase_action = settings.create_action("generator-use-uppercase")
        action_group.add_action(use_uppercase_action)

        use_lowercase_action = settings.create_action("generator-use-lowercase")
        action_group.add_action(use_lowercase_action)

        use_numbers_action = settings.create_action("generator-use-numbers")
        action_group.add_action(use_numbers_action)

        use_symbols_action = settings.create_action("generator-use-symbols")
        action_group.add_action(use_symbols_action)

        # Passphrase Generation
        settings.bind(
            "generator-words",
            self._generator_words_spin_row,
            "value",
            Gio.SettingsBindFlags.DEFAULT,
        )
        settings.bind(
            "generator-separator",
            self._generator_separator_entry,
            "text",
            Gio.SettingsBindFlags.DEFAULT,
        )

        # Security
        settings.bind(
            "database-lock-timeout",
            self._lockdb_spin_row,
            "value",
            Gio.SettingsBindFlags.DEFAULT,
        )
        settings.bind(
            "clear-clipboard",
            self._clearcb_spin_row,
            "value",
            Gio.SettingsBindFlags.DEFAULT,
        )

        self._clear_button.connect("clicked", self._on_settings_clear_recents_clicked)

        recents = RecentManager()
        if recents.is_empty():
            self._clear_button.props.sensitive = False

        # Unlock
        lock_on_session_lock = settings.create_action("lock-on-session-lock")
        action_group.add_action(lock_on_session_lock)

        remember_composite_key_action = settings.create_action("remember-composite-key")
        action_group.add_action(remember_composite_key_action)
        remember_composite_key_action.connect(
            "notify::state",
            self._on_remember_composite_key,
        )

        quick_unlock_action = settings.create_action("quick-unlock")
        action_group.add_action(quick_unlock_action)
        fingerprint_quick_unlock_action = settings.create_action(
            "fingerprint-quick-unlock",
        )
        action_group.add_action(fingerprint_quick_unlock_action)

        self.insert_action_group("settings", action_group)

    def _on_remember_composite_key(self, action, _param):
        if not action.props.state:
            config.set_last_used_key_provider({})

    def _on_settings_clear_recents_clicked(self, widget):
        recents = RecentManager()
        recents.purge_items()
        config.set_last_opened_database("")
        widget.set_sensitive(False)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gettext import gettext as _

from gi.repository import Gio

from gsecrets.recent_manager import RecentManager


class RecentFilesMenu:
    def __init__(self):
        """Recently opened files page menu.

        The `menu` attribute contains a GMenuModel to be used on popovers.
        """
        self.menu = Gio.Menu.new()
        self.section = Gio.Menu.new()
        self.is_empty = True

        recents = RecentManager()

        for item in reversed(recents):
            basename = item.get_basename()
            path = item.get_path()
            self.section.append(basename, f"win.open_database::{path}")
            self.is_empty = False

        self.menu.append_section(_("Recent Files"), self.section)
        self.menu.freeze()



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gi.repository import Adw, Gtk

from gsecrets import const


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/welcome_page.ui")
class WelcomePage(Adw.Bin):
    __gtype_name__ = "WelcomePage"

    _status_page = Gtk.Template.Child()

    def __init__(self):
        """Welcome Page widget."""
        super().__init__()

        self._status_page.set_icon_name(const.APP_ID)



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gettext import gettext as _

from gi.repository import Adw

from gsecrets.widgets.saving_conflict_dialog import SavingConflictDialog


class QuitConflictDialog(SavingConflictDialog):
    __gtype_name__ = "QuitConflictDialog"

    def __init__(self, window, db_manager, save_callback):
        super().__init__(window, db_manager, save_callback)

        # TRANSLATORS Dialog header for when there is a conflict.
        self.props.heading = _("Unsaved Changes Conflict")

        self.set_response_label("save", _("Save and Quit"))
        self.set_response_label("backup", _("Back up, Save, and Quit"))
        self.set_response_label("cancel", _("Don't Quit"))

        self.add_response("discard", _("_Quit Without Saving"))
        self.set_response_appearance("discard", Adw.ResponseAppearance.DESTRUCTIVE)
        self.connect("response::discard", self._on_response_discard)

    def _on_response_discard(self, _dialog, _response):
        self.window.force_close()



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import asyncio
import logging
from gettext import gettext as _

from gi.events import GLibEventLoopPolicy
from gi.repository import Adw, Gio, GLib, Gtk, GtkSource

from gsecrets import const
from gsecrets.recent_manager import RecentManager
from gsecrets.widgets.mod import load_widgets
from gsecrets.widgets.window import Window


class Application(Adw.Application):
    development_mode = const.IS_DEVEL
    application_id = const.APP_ID
    settings = Gio.Settings.new(application_id)

    def __init__(self, *args, **_kwargs):
        super().__init__(
            *args,
            application_id=self.application_id,
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
            resource_base_path="/org/gnome/World/Secrets",
            register_session=True,
        )

        asyncio.set_event_loop_policy(GLibEventLoopPolicy())

        # debug level logging option
        self.add_main_option(
            "debug",
            ord("d"),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            _("Enable debug logging"),
            None,
        )

    def do_startup(self):  # pylint: disable=arguments-differ
        Adw.Application.do_startup(self)

        GtkSource.init()

        Gtk.Window.set_default_icon_name(const.APP_ID)

        self.setup_actions()
        self.add_global_accelerators()

        load_widgets()

        recents = RecentManager()
        self.create_asyncio_task(recents.clean_non_existent())

    def do_open(self, gfile_list, _n_files, _hint):  # pylint: disable=arguments-differ
        for gfile in gfile_list:
            if not gfile.query_exists():
                logging.error("Error: File %s does not exist", gfile.get_path())
                if self.get_windows() is None:
                    self.quit()

            if self.is_safe_open(gfile.get_path()):
                logging.error("Error: Safe %s is already open", gfile.get_path())
            else:
                window = self.new_window()
                window.present()
                window.start_database_opening_routine(gfile.get_path())

    def new_window(self) -> Window:
        """Create a new window.

        The new window will be placed in its own group. This is done
        so that modal windows don't make other windows insensitive.
        """
        window_group = Gtk.WindowGroup()
        window = Window(application=self)

        window_group.add_window(window)
        return window

    def is_safe_open(self, filepath: str) -> bool:
        for window in self.get_windows():
            database = window.unlocked_db
            if not database:
                continue

            if database.database_manager.path == filepath:
                return True

        return False

    def do_handle_local_options(  # pylint: disable=arguments-differ
        self,
        options: GLib.VariantDict,
    ) -> int:
        """Handle cli arguments.

        :returns int: If you have handled your options and want to exit
        the process, return a non-negative option, 0 for success, and
        a positive value for failure. To continue, return -1 to let
        the default option processing continue.
        """
        #  convert GVariantDict -> GVariant -> dict
        options = options.end().unpack()

        # set up logging depending on the verbosity level
        loglevel = logging.INFO
        if self.development_mode or "debug" in options:
            loglevel = logging.DEBUG
            # Don't clutter our log output with debug msg of the
            # pykeepass module it is very noisy.
            pykeepass_logger = logging.getLogger("pykeepass")
            pykeepass_logger.setLevel(logging.INFO)

        logging.basicConfig(
            format="%(asctime)s | %(levelname)s | %(message)s",
            datefmt="%d-%m-%y %H:%M:%S",
            level=loglevel,
        )

        return -1

    def do_activate(self):  # pylint: disable=arguments-differ
        Adw.Application.do_activate(self)

        if window := self.get_active_window():
            window.present()
            return

        window = self.new_window()
        window.invoke_initial_screen()
        window.present()

    def setup_actions(self):
        quit_action = Gio.SimpleAction.new("quit", None)
        quit_action.connect("activate", self.on_quit_action)
        self.add_action(quit_action)

        new_window_action = Gio.SimpleAction.new("new-window", None)
        new_window_action.connect("activate", self.on_new_window_action)
        self.add_action(new_window_action)

    def on_quit_action(self, _action: Gio.Action, _param: GLib.Variant) -> None:
        for window in self.get_windows():
            window.close()

    def on_new_window_action(self, _action: Gio.Action, _param: GLib.Variant) -> None:
        window = self.new_window()
        window.invoke_initial_screen()
        window.present()

    def add_global_accelerators(self):
        self.set_accels_for_action("app.quit", ["<Control>q"])
        self.set_accels_for_action("app.new-window", ["<Control><Shift>n"])
        self.set_accels_for_action("win.settings", ["<Control>comma"])
        self.set_accels_for_action("win.open_database('')", ["<Control>o"])
        self.set_accels_for_action("win.new_database", ["<Control>n"])
        self.set_accels_for_action("win.db.save", ["<Control>s"])
        self.set_accels_for_action("win.db.search", ["<Control>f"])
        self.set_accels_for_action("win.db.lock", ["<Control>l"])
        self.set_accels_for_action("win.db.add_entry", ["<Control>e"])
        self.set_accels_for_action("win.db.add_group", ["<Control>p"])
        self.set_accels_for_action("win.go_back", ["Escape"])
        self.set_accels_for_action("window.close", ["<Control>w"])



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import typing
from gettext import gettext as _

from gi.repository import Adw, Gtk, GObject

if typing.TYPE_CHECKING:
    from collections.abc import Callable
    from gsecrets.database_manager import DatabaseManager
    from gsecrets.unlocked_database import UnlockedDatabase
    from pykeepass.attachment import Attachment


@Gtk.Template(resource_path="/org/gnome/World/Secrets/gtk/attachment_delete_dialog.ui")
class AttachmentDeleteDialog(Adw.AlertDialog):
    __gtype_name__ = "AttachmentDeleteDialog"

    def __init__(
        self,
        unlocked_database: UnlockedDatabase,
        attachment: Attachment,
        on_delete: Callable,
    ) -> None:
        """Dialog to confirm deleting an attachment."""
        super().__init__()

        self.__unlocked_database = unlocked_database
        self.__on_delete = on_delete
        self.__signal_handle = 0

        self.set_body(
            # TRANSLATORS The {filename} is a placeholder, must not be translated
            _(
                "The attachment “{filename}” will be deleted. "
                "This action cannot be undone."
            ).format(filename=attachment.filename)
        )

        self.__setup_signals()

    def __setup_signals(self):
        handle = self.__unlocked_database.database_manager.connect(
            "notify::locked",
            self.__on_locked,
        )
        self.__signal_handle = handle

    @Gtk.Template.Callback()
    def _on_delete(self, _dialog: Adw.AlertDialog, _response: str) -> None:
        self.__on_delete()

    def __on_locked(
        self, database_manager: DatabaseManager, _pspec: GObject.ParamSpec
    ) -> None:
        locked = database_manager.props.locked
        if locked:
            self.close()

    def do_closed(self) -> None:
        self.__unlocked_database.start_database_lock_timer()

        if handle := self.__signal_handle:
            self.__unlocked_database.database_manager.disconnect(handle)
            self.__signal_handle = 0



# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

from gettext import gettext as _

from gi.repository import GObject, Gtk

import gsecrets.config_manager as config
from gsecrets.passphrase_generator import Passphrase
from gsecrets.password_generator import generate as generate_pwd


@Gtk.Template(
    resource_path="/org/gnome/World/Secrets/gtk/password_generator_popover.ui",
)
class PasswordGeneratorPopover(Gtk.Popover):
    __gtype_name__ = "PasswordGeneratorPopover"

    _digit_spin_row = Gtk.Template.Child()
    _generate_button = Gtk.Template.Child()
    _high_letter_toggle_btn = Gtk.Template.Child()
    _low_letter_toggle_button = Gtk.Template.Child()
    _number_toggle_button = Gtk.Template.Child()
    _separator_entry = Gtk.Template.Child()
    _special_toggle_button = Gtk.Template.Child()
    _stack = Gtk.Template.Child()
    _words_spin_row = Gtk.Template.Child()

    def __init__(self):
        """Popover to generate a new password for an entry.

        :param UnlockedDatabase database_view: database view
        """
        super().__init__()

        use_lowercase = config.get_generator_use_lowercase()
        self._low_letter_toggle_button.set_active(use_lowercase)

        use_uppercase = config.get_generator_use_uppercase()
        self._high_letter_toggle_btn.set_active(use_uppercase)

        use_numbers = config.get_generator_use_numbers()
        self._number_toggle_button.set_active(use_numbers)

        use_symbols = config.get_generator_use_symbols()
        self._special_toggle_button.set_active(use_symbols)

        self._digit_spin_row.props.value = config.get_generator_length()

        self._separator_entry.props.text = config.get_generator_separator()

        delegate = self._separator_entry.get_delegate()
        delegate.props.max_length = 1

        self._words_spin_row.props.value = config.get_generator_words()

        self._low_letter_toggle_button.connect("toggled", self.on_toggled_callback)
        self._high_letter_toggle_btn.connect("toggled", self.on_toggled_callback)
        self._number_toggle_button.connect("toggled", self.on_toggled_callback)
        self._special_toggle_button.connect("toggled", self.on_toggled_callback)

        self.set_tooltips()

    def on_toggled_callback(self, _button):
        self.set_tooltips()

    def set_tooltips(self) -> None:
        use_uppercase: bool = self._high_letter_toggle_btn.props.active
        use_lowercase: bool = self._low_letter_toggle_button.props.active
        use_numbers: bool = self._number_toggle_button.props.active
        use_symbols: bool = self._special_toggle_button.props.active

        if use_lowercase:
            self._low_letter_toggle_button.set_tooltip_text(_("Include Lowercase"))
        else:
            self._low_letter_toggle_button.set_tooltip_text(_("Exclude Lowercase"))

        if use_uppercase:
            self._high_letter_toggle_btn.set_tooltip_text(_("Include Uppercase"))
        else:
            self._high_letter_toggle_btn.set_tooltip_text(_("Exclude Uppercase"))

        if use_numbers:
            self._number_toggle_button.set_tooltip_text(_("Include Numbers"))
        else:
            self._number_toggle_button.set_tooltip_text(_("Exclude Numbers"))

        if use_symbols:
            self._special_toggle_button.set_tooltip_text(_("Include Symbols"))
        else:
            self._special_toggle_button.set_tooltip_text(_("Exclude Symbols"))

    @Gtk.Template.Callback()
    def _on_generate_button_clicked(self, _button: Gtk.Button) -> None:
        if self._stack.get_visible_child_name() == "password":
            use_uppercase: bool = self._high_letter_toggle_btn.props.active
            use_lowercase: bool = self._low_letter_toggle_button.props.active
            use_numbers: bool = self._number_toggle_button.props.active
            use_symbols: bool = self._special_toggle_button.props.active

            length = int(self._digit_spin_row.props.value)
            pass_text: str = generate_pwd(
                length,
                use_uppercase,
                use_lowercase,
                use_numbers,
                use_symbols,
            )
            self.emit("generated", pass_text)
        else:
            separator: str = self._separator_entry.props.text
            words = int(self._words_spin_row.props.value)
            passphrase_generator = Passphrase()
            passphrase_generator.connect("generated", self.on_passphrase_generated)
            passphrase_generator.generate(words, separator)

    @GObject.Signal(arg_types=(str,))
    def generated(self, _password):
        return

    def on_passphrase_generated(self, _passphrase_gen, passphrase):
        self.emit("generated", passphrase)



# SPDX-License-Identifier: GPL-3.0-only
import logging
from gettext import gettext as _

from gi.repository import Adw, Gio, GLib, Gtk

import gsecrets.config_manager as config
from gsecrets.database_manager import DatabaseManager
from gsecrets.provider.base_provider import BaseProvider
from gsecrets.utils import (
    KeyFileFilter,
    LazyValue,
    generate_keyfile_async,
    generate_keyfile_finish,
)


class FileProvider(BaseProvider):
    # pylint: disable=too-many-instance-attributes
    def __init__(self, window):
        super().__init__()

        self.raw_key = None
        self.keyfile_path = None
        self.keyfile_hash = ""
        self.unlock_row = None
        self.unlock_clear_button = None
        self.create_row = None
        self.create_clear_button = None
        self.window = window

    def create_unlock_widget(self, database_manager: DatabaseManager) -> Gtk.Widget:
        self.unlock_row = Adw.ActionRow()
        self.unlock_row.set_title(_("Keyfile"))
        if self.keyfile_path:
            self.unlock_row.set_subtitle(self.keyfile_path)

        button = Gtk.Button()
        button.set_valign(Gtk.Align.CENTER)
        button.add_css_class("flat")
        button.set_icon_name("folder-open-symbolic")
        button.set_tooltip_text(_("Select Keyfile"))
        button.connect("clicked", self._on_keyfile_button_clicked)
        self.unlock_row.add_suffix(button)

        self.unlock_clear_button = Gtk.Button()
        self.unlock_clear_button.set_visible(False)
        self.unlock_clear_button.set_valign(Gtk.Align.CENTER)
        self.unlock_clear_button.add_css_class("flat")
        self.unlock_clear_button.set_icon_name("edit-delete-symbolic")
        self.unlock_clear_button.set_tooltip_text(_("Clear Keyfile"))
        self.unlock_clear_button.connect("clicked", self._on_unlock_clear_keyfile)
        self.unlock_row.add_suffix(self.unlock_clear_button)

        pairs = config.get_provider_config(database_manager.path, "FileProvider")

        self._set_keyfile(None)
        if pairs and pairs["url"] is not None:
            keyfile = Gio.File.new_for_path(pairs["url"])
            if keyfile.query_exists():
                self._set_keyfile(keyfile)

        return self.unlock_row

    def _on_unlock_clear_keyfile(self, _widget):
        self.keyfile_path = ""
        self.keyfile_hash = ""
        self.raw_key = None

        self.unlock_row.set_subtitle(_("Select keyfile"))
        self.unlock_clear_button.set_visible(False)

    def _on_keyfile_button_clicked(self, _widget: Gtk.Widget) -> None:
        """Cb invoked when we unlock a database via keyfile."""
        filters = Gio.ListStore.new(Gtk.FileFilter)
        filters.append(KeyFileFilter().file_filter)

        dialog = Gtk.FileDialog.new()
        dialog.props.filters = filters
        dialog.props.title = _("Select Keyfile")

        dialog.open(self.window, None, self._on_dialog_response)

    def _on_dialog_response(self, dialog: Gtk.Dialog, result: int) -> None:
        try:
            keyfile = dialog.open_finish(result)
        except GLib.Error as err:
            if not err.matches(Gtk.DialogError.quark(), Gtk.DialogError.DISMISSED):
                logging.exception("Could not open file")
        else:
            self._set_keyfile(keyfile)

    def _set_keyfile(self, keyfile: Gio.File) -> None:
        if keyfile is None:
            self._on_unlock_clear_keyfile(None)
            return

        self.keyfile_path = keyfile.get_path()
        keyfile.load_bytes_async(None, self.load_keyfile_callback)

        logging.debug("Keyfile selected: %s", self.keyfile_path)

    def load_keyfile_callback(self, keyfile, result):
        try:
            key, _etag = keyfile.load_bytes_finish(result)
        except GLib.Error:
            logging.exception("Could not set keyfile hash")
            self.window.send_notification(_("Could not load keyfile"))
        else:
            keyfile_hash = GLib.compute_checksum_for_bytes(GLib.ChecksumType.SHA1, key)
            self.raw_key = key.get_data()
            file_path = keyfile.get_path()
            basename = keyfile.get_basename()
            if keyfile_hash:
                self.keyfile_hash = keyfile_hash
                self.keyfile_path = file_path

            if self.unlock_row:
                self.unlock_row.set_subtitle(basename)
                self.unlock_clear_button.set_visible(True)

            if self.create_row:
                self.create_row.set_subtitle(basename)
                self.create_row.set_visible(True)
        finally:
            if self.unlock_row:
                self.unlock_row.props.sensitive = True

    def create_database_row(self):
        self.create_row = Adw.ActionRow()
        self.create_row.set_title(_("Keyfile"))

        generate_button = Gtk.Button()
        generate_button.set_icon_name("document-new-symbolic")
        generate_button.set_tooltip_text(_("Generate Keyfile"))
        generate_button.connect("clicked", self.on_generate_keyfile_button_clicked)
        generate_button.add_css_class("flat")
        generate_button.set_valign(Gtk.Align.CENTER)
        self.create_row.add_suffix(generate_button)

        create_clear_button = Gtk.Button()
        create_clear_button.set_tooltip_text(_("Clear Keyfile"))
        create_clear_button.set_icon_name("edit-delete-symbolic")
        create_clear_button.connect("clicked", self._on_create_clear_keyfile)
        create_clear_button.add_css_class("flat")
        create_clear_button.set_valign(Gtk.Align.CENTER)
        self.create_clear_button = create_clear_button
        self.create_row.add_suffix(create_clear_button)

        self._on_create_clear_keyfile(None)

        return self.create_row

    def _on_create_clear_keyfile(self, _widget):
        self.keyfile_path = ""
        self.keyfile_hash = ""
        self.raw_key = None

        self.create_row.set_subtitle(_("Use a file to increase security"))
        self.create_clear_button.set_visible(False)

    @property
    def available(self):
        return True

    def _on_filechooser_response(self, dialog, result):
        try:
            keyfile = dialog.save_finish(result)
        except GLib.Error as err:
            if not err.matches(Gtk.DialogError.quark(), Gtk.DialogError.DISMISSED):
                logging.exception("Could not save file")
        else:
            keyfile_path = keyfile.get_path()
            logging.debug("New keyfile location: %s", keyfile_path)

            def callback(_gfile, result):
                try:
                    _res, _hash = generate_keyfile_finish(result)
                except GLib.Error:
                    logging.exception("Could not create keyfile")
                    self.window.send_notification(_("Could not create keyfile"))
                else:
                    self.create_clear_button.set_visible(True)
                    keyfile.load_bytes_async(None, self.load_keyfile_callback)

            generate_keyfile_async(keyfile, callback)

    def on_generate_keyfile_button_clicked(self, _widget: Gtk.Button) -> None:
        """Cb invoked when we create a new keyfile for a newly created Safe."""
        filters = Gio.ListStore.new(Gtk.FileFilter)
        filters.append(KeyFileFilter().file_filter)

        dialog = Gtk.FileDialog.new()
        dialog.props.filters = filters
        dialog.props.title = _("Generate Keyfile")
        dialog.props.accept_label = _("_Generate")
        dialog.props.initial_name = _("Keyfile")

        dialog.save(self.window, None, self._on_filechooser_response)

    def config(self):
        ret = {}

        if self.keyfile_path:
            ret["url"] = self.keyfile_path

        return ret

    def clear_input_fields(self):
        self._on_create_clear_keyfile(None)

    def generate_key(self, _salt: LazyValue[bytes]) -> bool:
        return self.raw_key is not None
